@@ -1,5 +1,5 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set sw=4 sts=4 ts=4 expandtab: */
+/* vim: set sw=4 sts=4 expandtab: */
 /*
    rsvg.c: SAX-based renderer for SVG files into a GdkPixbuf.
 
@@ -27,9 +27,8 @@
 #include "config.h"
 #define _GNU_SOURCE 1
 
-#include "rsvg.h"
-#include "rsvg-compat.h"
 #include "rsvg-private.h"
+#include "rsvg-compat.h"
 #include "rsvg-css.h"
 #include "rsvg-styles.h"
 #include "rsvg-shapes.h"
@@ -182,7 +181,20 @@ rsvg_start_style (RsvgHandle * ctx, RsvgPropertyBag *atts)
     handler->ctx = ctx;
 
     handler->style = g_string_new (NULL);
-    handler->is_text_css = type && g_ascii_strcasecmp (type, "text/css") == 0;
+
+    /* FIXME: See these:
+     *
+     * https://www.w3.org/TR/SVG/styling.html#StyleElementTypeAttribute
+     * https://www.w3.org/TR/SVG/styling.html#ContentStyleTypeAttribute
+     *
+     * If the "type" attribute is not present, we should fallback to the
+     * "contentStyleType" attribute of the svg element, which in turn
+     * defaults to "text/css".
+     *
+     * See where is_text_css is used to see where we parse the contents
+     * of the style element.
+     */
+    handler->is_text_css = (type == NULL) || (g_ascii_strcasecmp (type, "text/css") == 0);
 
     handler->parent = (RsvgSaxHandlerDefs *) ctx->priv->handler;
     ctx->priv->handler = &handler->super;
@@ -194,7 +206,7 @@ add_node_to_handle (RsvgHandle *ctx, RsvgNode *node)
     g_assert (ctx != NULL);
     g_assert (node != NULL);
 
-    g_ptr_array_add (ctx->priv->all_nodes, node);
+    g_ptr_array_add (ctx->priv->all_nodes, rsvg_node_ref (node));
 }
 
 static void
@@ -239,7 +251,7 @@ free_element_name_stack (RsvgHandle *ctx)
     ctx->priv->element_name_stack = NULL;
 }
 
-typedef RsvgNode *(* CreateNodeFn) (const char *element_name);
+typedef RsvgNode *(* CreateNodeFn) (const char *element_name, RsvgNode *parent);
 
 typedef struct {
     const char   *element_name;
@@ -252,7 +264,7 @@ typedef struct {
  * Lines in comments are elements that we don't support.
  */
 static const NodeCreator node_creators[] = {
-    { "a",                   TRUE,  rsvg_new_group },    /* treat anchors as groups for now */
+    { "a",                   TRUE,  rsvg_node_group_new },    /* treat anchors as groups for now */
     /* "altGlyph",           TRUE,  */
     /* "altGlyphDef",        FALSE, */
     /* "altGlyphItem",       FALSE, */
@@ -260,14 +272,14 @@ static const NodeCreator node_creators[] = {
     /* "animateColor",       FALSE, */
     /* "animateMotion",      FALSE, */
     /* "animateTransform",   FALSE, */
-    { "circle",              TRUE,  rsvg_new_circle },
+    { "circle",              TRUE,  rsvg_node_circle_new },
     { "clipPath",            TRUE,  rsvg_new_clip_path },
     /* "color-profile",      FALSE, */
-    { "conicalGradient",     TRUE,  rsvg_new_radial_gradient },
+    { "conicalGradient",     TRUE,  rsvg_node_radial_gradient_new },
     /* "cursor",             FALSE, */
-    { "defs",                TRUE,  rsvg_new_defs },
+    { "defs",                TRUE,  rsvg_node_defs_new },
     /* "desc",               TRUE,  */
-    { "ellipse",             TRUE,  rsvg_new_ellipse },
+    { "ellipse",             TRUE,  rsvg_node_ellipse_new },
     { "feBlend",             TRUE,  rsvg_new_filter_primitive_blend },
     { "feColorMatrix",       TRUE,  rsvg_new_filter_primitive_color_matrix },
     { "feComponentTransfer", TRUE,  rsvg_new_filter_primitive_component_transfer },
@@ -300,48 +312,49 @@ static const NodeCreator node_creators[] = {
     /* "font-face-src",      FALSE, */
     /* "font-face-uri",      FALSE, */
     /* "foreignObject",      TRUE,  */
-    { "g",                   TRUE,  rsvg_new_group },
+    { "g",                   TRUE,  rsvg_node_group_new },
     /* "glyph",              TRUE,  */
     /* "glyphRef",           TRUE,  */
     /* "hkern",              FALSE, */
     { "image",               TRUE,  rsvg_new_image },
-    { "line",                TRUE,  rsvg_new_line },
-    { "linearGradient",      TRUE,  rsvg_new_linear_gradient },
-    { "marker",              TRUE,  rsvg_new_marker },
+    { "line",                TRUE,  rsvg_node_line_new },
+    { "linearGradient",      TRUE,  rsvg_node_linear_gradient_new },
+    { "marker",              TRUE,  rsvg_node_marker_new },
     { "mask",                TRUE,  rsvg_new_mask },
     /* "metadata",           FALSE, */
     /* "missing-glyph",      TRUE,  */
     /* "mpath"               FALSE, */
-    { "multiImage",          FALSE, rsvg_new_switch }, /* hack to make multiImage sort-of work */
-    { "path",                TRUE,  rsvg_new_path },
-    { "pattern",             TRUE,  rsvg_new_pattern },
-    { "polygon",             TRUE,  rsvg_new_polygon },
-    { "polyline",            TRUE,  rsvg_new_polyline },
-    { "radialGradient",      TRUE,  rsvg_new_radial_gradient },
-    { "rect",                TRUE,  rsvg_new_rect },
+    { "multiImage",          FALSE, rsvg_node_switch_new }, /* hack to make multiImage sort-of work */
+    { "path",                TRUE,  rsvg_node_path_new },
+    { "pattern",             TRUE,  rsvg_node_pattern_new },
+    { "polygon",             TRUE,  rsvg_node_polygon_new },
+    { "polyline",            TRUE,  rsvg_node_polyline_new },
+    { "radialGradient",      TRUE,  rsvg_node_radial_gradient_new },
+    { "rect",                TRUE,  rsvg_node_rect_new },
     /* "script",             FALSE, */
     /* "set",                FALSE, */
-    { "stop",                TRUE,  rsvg_new_stop },
+    { "stop",                TRUE,  rsvg_node_stop_new },
     /* "style",              FALSE, */
-    { "subImage",            FALSE, rsvg_new_group },
+    { "subImage",            FALSE, rsvg_node_group_new },
     { "subImageRef",         FALSE, rsvg_new_image },
-    { "svg",                 TRUE,  rsvg_new_svg },
-    { "switch",              TRUE,  rsvg_new_switch },
-    { "symbol",              TRUE,  rsvg_new_symbol },
+    { "svg",                 TRUE,  rsvg_node_svg_new },
+    { "switch",              TRUE,  rsvg_node_switch_new },
+    { "symbol",              TRUE,  rsvg_node_symbol_new },
     { "text",                TRUE,  rsvg_new_text },
     /* "textPath",           TRUE,  */
     /* "title",              TRUE,  */
     { "tref",                TRUE,  rsvg_new_tref },
     { "tspan",               TRUE,  rsvg_new_tspan },
-    { "use",                 TRUE,  rsvg_new_use },
+    { "use",                 TRUE,  rsvg_node_use_new },
     /* "view",               FALSE, */
     /* "vkern",              FALSE, */
 };
 
-/* hack for bug 401115. whenever we encounter a node we don't understand, push it into a group.
- * this will allow us to handle things like conditionals properly.
+/* Whenever we encounter a node we don't understand, represent it as a defs.
+ * This is like a group, but it doesn't do any rendering of children.  The
+ * effect is that we will ignore all children of unknown elements.
  */
-static const NodeCreator default_node_creator = { NULL, TRUE, rsvg_new_group };
+static const NodeCreator default_node_creator = { NULL, TRUE, rsvg_node_defs_new };
 
 /* Used from bsearch() */
 static int
@@ -382,13 +395,13 @@ node_set_atts (RsvgNode * node, RsvgHandle * ctx, const NodeCreator *creator, Rs
         const char *id;
         const char *klazz;
 
-        node->vtable->set_atts (node, ctx, atts);
+        rsvg_node_set_atts (node, ctx, atts);
 
         /* The "svg" node is special; it will load its id/class
          * attributes until the end, when rsvg_end_element() calls
          * _rsvg_node_svg_apply_atts()
          */
-        if (rsvg_node_type (node) != RSVG_NODE_TYPE_SVG) {
+        if (rsvg_node_get_type (node) != RSVG_NODE_TYPE_SVG) {
             id = rsvg_property_bag_lookup (atts, "id");
 
             if (creator->supports_class_attribute)
@@ -396,7 +409,7 @@ node_set_atts (RsvgNode * node, RsvgHandle * ctx, const NodeCreator *creator, Rs
             else
                 klazz = NULL;
 
-            rsvg_parse_style_attrs (ctx, rsvg_node_get_state (node), creator->element_name, klazz, id, atts);
+            rsvg_parse_style_attrs (ctx, node, creator->element_name, klazz, id, atts);
         }
     }
 }
@@ -410,63 +423,28 @@ rsvg_standard_element_start (RsvgHandle * ctx, const char *name, RsvgPropertyBag
     creator = get_node_creator_for_element_name (name);
     g_assert (creator != NULL && creator->create_fn != NULL);
 
-    newnode = creator->create_fn (name);
+    newnode = creator->create_fn (name, ctx->priv->currentnode);
+    g_assert (newnode != NULL);
 
-    if (newnode) {
-        g_assert (rsvg_node_type (newnode) != RSVG_NODE_TYPE_INVALID);
+    g_assert (rsvg_node_get_type (newnode) != RSVG_NODE_TYPE_INVALID);
 
-        push_element_name (ctx, name);
+    push_element_name (ctx, name);
 
-        add_node_to_handle (ctx, newnode);
-        register_node_in_defs (ctx, newnode, atts);
+    add_node_to_handle (ctx, newnode);
+    register_node_in_defs (ctx, newnode, atts);
 
-        if (ctx->priv->currentnode) {
-            rsvg_node_add_child (ctx->priv->currentnode, newnode);
-        } else if (rsvg_node_type (newnode) == RSVG_NODE_TYPE_SVG) {
-            ctx->priv->treebase = newnode;
-        }
-
-        ctx->priv->currentnode = newnode;
-
-        node_set_atts (newnode, ctx, creator, atts);
+    if (ctx->priv->currentnode) {
+        rsvg_node_add_child (ctx->priv->currentnode, newnode);
+        ctx->priv->currentnode = rsvg_node_unref (ctx->priv->currentnode);
+    } else if (rsvg_node_get_type (newnode) == RSVG_NODE_TYPE_SVG) {
+        ctx->priv->treebase = rsvg_node_ref (newnode);
     }
-}
 
-RsvgNodeType
-rsvg_node_type (RsvgNode *node)
-{
-    return node->type;
-}
+    ctx->priv->currentnode = rsvg_node_ref (newnode);
 
-RsvgState *
-rsvg_node_get_state (RsvgNode *node)
-{
-    return node->state;
-}
+    node_set_atts (newnode, ctx, creator, atts);
 
-RsvgNode *
-rsvg_node_get_parent (RsvgNode *node)
-{
-    return node->parent;
-}
-
-void
-rsvg_node_foreach_child (RsvgNode *node, RsvgNodeForeachChildFn fn, gpointer data)
-{
-    guint len;
-    guint i;
-
-    len = node->children->len;
-
-    for (i = 0; i < len; i++) {
-        RsvgNode *child;
-        gboolean next;
-
-        child = g_ptr_array_index (node->children, i);
-        next = fn (child, data);
-        if (!next)
-            break;
-    }
+    newnode = rsvg_node_unref (newnode);
 }
 
 /* extra (title, desc, metadata) */
@@ -544,7 +522,7 @@ rsvg_start_extra (RsvgHandle * ctx,
      * This isn't quite the correct behavior - any graphics
      * element may contain a <extra> element.
      */
-    do_care = treebase != NULL && treebase == currentnode;
+    do_care = treebase != NULL && rsvg_node_is_same (treebase, currentnode);
 
     handler->super.free = rsvg_extra_handler_free;
     handler->super.characters = rsvg_extra_handler_characters;
@@ -756,7 +734,6 @@ rsvg_start_xinclude (RsvgHandle * ctx, RsvgPropertyBag * atts)
         /* xml */
         GInputStream *stream;
         GError *err = NULL;
-        xmlDocPtr xml_doc;
         xmlParserCtxtPtr xml_parser;
         xmlParserInputBufferPtr buffer;
         xmlParserInputPtr input;
@@ -776,16 +753,14 @@ rsvg_start_xinclude (RsvgHandle * ctx, RsvgPropertyBag * atts)
         if (xmlPushInput (xml_parser, input) < 0) {
             g_clear_error (&err);
             xmlFreeInputStream (input);
-            xmlFreeParserCtxt (xml_parser);
+
+            xml_parser = rsvg_free_xml_parser_and_doc (xml_parser);
             goto fallback;
         }
 
         (void) xmlParseDocument (xml_parser);
 
-        xml_doc = xml_parser->myDoc;
-        xmlFreeParserCtxt (xml_parser);
-        if (xml_doc)
-            xmlFreeDoc (xml_doc);
+        xml_parser = rsvg_free_xml_parser_and_doc (xml_parser);
 
         g_clear_error (&err);
     }
@@ -867,37 +842,51 @@ rsvg_end_element (void *data, const xmlChar * xmlname)
         }
 
         if (ctx->priv->currentnode && topmost_element_name_is (ctx, name)) {
-            ctx->priv->currentnode = rsvg_node_get_parent (ctx->priv->currentnode);
+            RsvgNode *parent;
+
+            parent = rsvg_node_get_parent (ctx->priv->currentnode);
+            ctx->priv->currentnode = rsvg_node_unref (ctx->priv->currentnode);
+            ctx->priv->currentnode = parent;
             pop_element_name (ctx);
         }
 
         /* FIXMEchpe: shouldn't this check that currentnode == treebase or sth like that? */
-        if (ctx->priv->treebase && !strcmp (name, "svg"))
-            _rsvg_node_svg_apply_atts ((RsvgNodeSvg *)ctx->priv->treebase, ctx);
+        if (ctx->priv->treebase && !strcmp (name, "svg")) {
+            g_assert (rsvg_node_get_type (ctx->priv->treebase) == RSVG_NODE_TYPE_SVG);
+            rsvg_node_svg_apply_atts (ctx->priv->treebase, ctx);
+        }
     }
 }
 
 static void
-_rsvg_node_chars_free (RsvgNode * node)
+rsvg_node_chars_set_atts (RsvgNode *node, gpointer impl, RsvgHandle *handle, RsvgPropertyBag * atts)
 {
-    RsvgNodeChars *self = (RsvgNodeChars *) node;
-    g_string_free (self->contents, TRUE);
-    _rsvg_node_free (node);
+    /* nothing */
 }
 
-static RsvgNodeChars *
+static void
+rsvg_node_chars_draw (RsvgNode *node, gpointer impl, RsvgDrawingCtx *ctx, int dominate)
+{
+    /* nothing */
+}
+
+static void
+rsvg_node_chars_free (gpointer impl)
+{
+    RsvgNodeChars *self = impl;
+    g_string_free (self->contents, TRUE);
+    g_free (self);
+}
+
+static RsvgNode *
 rsvg_new_node_chars (const char *text,
-                     int len)
+                     int len,
+                     RsvgNode *parent)
 {
     RsvgNodeChars *self;
-    RsvgNodeVtable vtable = {
-        _rsvg_node_chars_free,
-        NULL,
-        NULL
-    };
+    RsvgState *state;
 
-    self = g_new (RsvgNodeChars, 1);
-    _rsvg_node_init (&self->super, RSVG_NODE_TYPE_CHARS, &vtable);
+    self = g_new0 (RsvgNodeChars, 1);
 
     if (!g_utf8_validate (text, len, NULL)) {
         char *utf8;
@@ -908,9 +897,16 @@ rsvg_new_node_chars (const char *text,
         self->contents = g_string_new_len (text, len);
     }
 
-    self->super.state->cond_true = FALSE;
+    state = rsvg_state_new ();
+    state->cond_true = FALSE;
 
-    return self;
+    return rsvg_rust_cnode_new (RSVG_NODE_TYPE_CHARS,
+                                parent,
+                                state,
+                                self,
+                                rsvg_node_chars_set_atts,
+                                rsvg_node_chars_draw,
+                                rsvg_node_chars_free);
 }
 
 static gboolean
@@ -920,10 +916,10 @@ find_last_chars_node (RsvgNode *node, gpointer data)
 
     dest = data;
 
-    if (rsvg_node_type (node) == RSVG_NODE_TYPE_CHARS) {
-        *dest = node;
-    } else if (rsvg_node_type (node) == RSVG_NODE_TYPE_TSPAN) {
-        *dest = NULL;
+    if (rsvg_node_get_type (node) == RSVG_NODE_TYPE_CHARS) {
+        *dest = rsvg_node_ref (node);
+    } else if (rsvg_node_get_type (node) == RSVG_NODE_TYPE_TSPAN) {
+        *dest = rsvg_node_unref (*dest); /* Discard the last chars node we found */
     }
 
     return TRUE;
@@ -932,22 +928,27 @@ find_last_chars_node (RsvgNode *node, gpointer data)
 static void
 rsvg_characters_impl (RsvgHandle * ctx, const xmlChar * ch, int len)
 {
-    RsvgNodeChars *self;
+    RsvgNode *node;
 
     if (!ch || !len)
         return;
 
     if (ctx->priv->currentnode) {
-        RsvgNodeType type = rsvg_node_type (ctx->priv->currentnode);
+        RsvgNodeType type = rsvg_node_get_type (ctx->priv->currentnode);
         if (type == RSVG_NODE_TYPE_TSPAN || type == RSVG_NODE_TYPE_TEXT) {
+            RsvgNodeChars *self;
+
             /* find the last CHARS node in the text or tspan node, so that we
                can coalesce the text, and thus avoid screwing up the Pango layouts */
-            self = NULL;
+            node = NULL;
             rsvg_node_foreach_child (ctx->priv->currentnode,
                                      find_last_chars_node,
-                                     &self);
+                                     &node);
+
+            if (node) {
+                g_assert (rsvg_node_get_type (node) == RSVG_NODE_TYPE_CHARS);
+                self = rsvg_rust_cnode_get_impl (node);
 
-            if (self != NULL) {
                 if (!g_utf8_validate ((char *) ch, len, NULL)) {
                     char *utf8;
                     utf8 = rsvg_make_valid_utf8 ((char *) ch, len);
@@ -957,17 +958,20 @@ rsvg_characters_impl (RsvgHandle * ctx, const xmlChar * ch, int len)
                     g_string_append_len (self->contents, (char *)ch, len);
                 }
 
+                node = rsvg_node_unref (node);
                 return;
             }
         }
     }
 
-    self = rsvg_new_node_chars ((char *) ch, len);
+    node = rsvg_new_node_chars ((char *) ch, len, ctx->priv->currentnode);
 
-    add_node_to_handle (ctx, (RsvgNode *) self);
+    add_node_to_handle (ctx, node);
 
     if (ctx->priv->currentnode)
-        rsvg_node_add_child (ctx->priv->currentnode, (RsvgNode *) self);
+        rsvg_node_add_child (ctx->priv->currentnode, node);
+
+    node = rsvg_node_unref (node);
 }
 
 static void
@@ -1355,31 +1359,23 @@ rsvg_handle_close_impl (RsvgHandle * handle, GError ** error)
 {
     GError *real_error = NULL;
 
-	handle->priv->is_closed = TRUE;
-
     handle->priv->error = &real_error;
 
     if (handle->priv->ctxt != NULL) {
-        xmlDocPtr xml_doc;
         int result;
 
-        xml_doc = handle->priv->ctxt->myDoc;
-
         result = xmlParseChunk (handle->priv->ctxt, "", 0, TRUE);
         if (result != 0) {
             rsvg_set_error (error, handle->priv->ctxt);
-            xmlFreeParserCtxt (handle->priv->ctxt);
-            xmlFreeDoc (xml_doc);
+            handle->priv->ctxt = rsvg_free_xml_parser_and_doc (handle->priv->ctxt);
             return FALSE;
         }
 
-        xmlFreeParserCtxt (handle->priv->ctxt);
-        xmlFreeDoc (xml_doc);
+        handle->priv->ctxt = rsvg_free_xml_parser_and_doc (handle->priv->ctxt);
     }
 
     free_element_name_stack (handle);
 
-    handle->priv->finished = TRUE;
     handle->priv->error = NULL;
 
     if (real_error != NULL) {
@@ -1397,8 +1393,7 @@ rsvg_drawing_ctx_free (RsvgDrawingCtx * handle)
 
     rsvg_state_free_all (handle->state);
 
-	/* the drawsub stack's nodes are owned by the ->defs */
-	g_slist_free (handle->drawsub_stack);
+	g_slist_free_full (handle->drawsub_stack, (GDestroyNotify) rsvg_node_unref);
 
     g_warn_if_fail (handle->acquired_nodes == NULL);
     g_slist_free (handle->acquired_nodes);
@@ -1526,9 +1521,10 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
     cairo_t *cr;
     cairo_surface_t *target;
     RsvgDrawingCtx *draw;
-    RsvgNodeSvg *root = NULL;
     RsvgNode *sself = NULL;
     RsvgBbox bbox;
+    RsvgLength root_width, root_height;
+    RsvgViewBox root_vbox;
 
     gboolean handle_subelement = TRUE;
 
@@ -1540,25 +1536,28 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
     if (id && *id) {
         sself = rsvg_defs_lookup (handle->priv->defs, id);
 
-        if (sself == handle->priv->treebase)
+        if (rsvg_node_is_same (sself, handle->priv->treebase))
             id = NULL;
-    }
-    else
+    } else {
         sself = handle->priv->treebase;
+    }
 
     if (!sself && id)
         return FALSE;
 
-    root = (RsvgNodeSvg *) handle->priv->treebase;
-
-    if (!root)
+    if (!handle->priv->treebase)
         return FALSE;
 
+    g_assert (rsvg_node_get_type (handle->priv->treebase) == RSVG_NODE_TYPE_SVG);
+
     bbox.rect.x = bbox.rect.y = 0;
     bbox.rect.width = bbox.rect.height = 1;
 
+    rsvg_node_svg_get_size (handle->priv->treebase, &root_width, &root_height);
+    root_vbox = rsvg_node_svg_get_view_box (handle->priv->treebase);
+
     if (!id) {
-        if ((root->w.unit == LENGTH_UNIT_PERCENT || root->h.unit == LENGTH_UNIT_PERCENT) && !root->vbox.active)
+        if ((root_width.unit == LENGTH_UNIT_PERCENT || root_height.unit == LENGTH_UNIT_PERCENT) && !root_vbox.active)
             handle_subelement = TRUE;
         else
             handle_subelement = FALSE;
@@ -1578,12 +1577,10 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
             return FALSE;
         }
 
-        while (sself != NULL) {
-            draw->drawsub_stack = g_slist_prepend (draw->drawsub_stack, sself);
-            sself = rsvg_node_get_parent (sself);
-        }
+        g_assert (sself != NULL);
+        rsvg_drawing_ctx_add_node_and_ancestors_to_stack (draw, sself);
 
-        rsvg_node_draw (handle->priv->treebase, draw, 0);
+        rsvg_drawing_ctx_draw_node_from_stack (draw, handle->priv->treebase, 0);
         bbox = RSVG_CAIRO_RENDER (draw->render)->bbox;
 
         rsvg_drawing_ctx_free (draw);
@@ -1593,12 +1590,12 @@ rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimensi
         dimension_data->width = bbox.rect.width;
         dimension_data->height = bbox.rect.height;
     } else {
-        bbox.rect.width = root->vbox.rect.width;
-        bbox.rect.height = root->vbox.rect.height;
+        bbox.rect.width = root_vbox.rect.width;
+        bbox.rect.height = root_vbox.rect.height;
 
-        dimension_data->width = (int) (rsvg_length_hand_normalize (&root->w, handle->priv->dpi_x,
+        dimension_data->width = (int) (rsvg_length_hand_normalize (&root_width, handle->priv->dpi_x,
                                                                    bbox.rect.width, 12) + 0.5);
-        dimension_data->height = (int) (rsvg_length_hand_normalize (&root->h, handle->priv->dpi_y,
+        dimension_data->height = (int) (rsvg_length_hand_normalize (&root_height, handle->priv->dpi_y,
                                                                     bbox.rect.height, 12) + 0.5);
     }
 
@@ -1629,7 +1626,6 @@ gboolean
 rsvg_handle_get_position_sub (RsvgHandle * handle, RsvgPositionData * position_data, const char *id)
 {
     RsvgDrawingCtx		*draw;
-    RsvgNodeSvg			*root;
     RsvgNode			*node;
     RsvgBbox			 bbox;
     RsvgDimensionData    dimension_data;
@@ -1640,6 +1636,9 @@ rsvg_handle_get_position_sub (RsvgHandle * handle, RsvgPositionData * position_d
     g_return_val_if_fail (handle, FALSE);
     g_return_val_if_fail (position_data, FALSE);
 
+    if (!handle->priv->treebase)
+        return FALSE;
+
     /* Short-cut when no id is given. */
     if (NULL == id || '\0' == *id) {
         position_data->x = 0;
@@ -1653,29 +1652,23 @@ rsvg_handle_get_position_sub (RsvgHandle * handle, RsvgPositionData * position_d
     node = rsvg_defs_lookup (handle->priv->defs, id);
     if (!node) {
         return FALSE;
-    } else if (node == handle->priv->treebase) {
+    } else if (rsvg_node_is_same (node, handle->priv->treebase)) {
         /* Root node. */
         position_data->x = 0;
         position_data->y = 0;
         return TRUE;
     }
 
-    root = (RsvgNodeSvg *) handle->priv->treebase;
-    if (!root)
-        return FALSE;
-
     target = cairo_image_surface_create (CAIRO_FORMAT_RGB24, 1, 1);
     cr = cairo_create  (target);
     draw = rsvg_cairo_new_drawing_ctx (cr, handle);
     if (!draw)
         goto bail;
 
-    while (node != NULL) {
-        draw->drawsub_stack = g_slist_prepend (draw->drawsub_stack, node);
-        node = rsvg_node_get_parent (node);
-    }
+    g_assert (node != NULL);
+    rsvg_drawing_ctx_add_node_and_ancestors_to_stack (draw, node);
 
-    rsvg_node_draw (handle->priv->treebase, draw, 0);
+    rsvg_drawing_ctx_draw_node_from_stack (draw, handle->priv->treebase, 0);
     bbox = RSVG_CAIRO_RENDER (draw->render)->bbox;
 
     rsvg_drawing_ctx_free (draw);
@@ -1890,10 +1883,13 @@ rsvg_handle_write (RsvgHandle * handle, const guchar * buf, gsize count, GError
     rsvg_return_val_if_fail (handle, FALSE, error);
     priv = handle->priv;
 
-    rsvg_return_val_if_fail (!priv->is_closed, FALSE, error);
+    rsvg_return_val_if_fail (priv->state == RSVG_HANDLE_STATE_START
+                             || priv->state == RSVG_HANDLE_STATE_READING,
+                             FALSE,
+                             error);
 
-    if (priv->first_write) {
-        priv->first_write = FALSE;
+    if (priv->state == RSVG_HANDLE_STATE_START) {
+        priv->state = RSVG_HANDLE_STATE_READING;
 
         /* test for GZ marker. todo: store the first 2 bytes in the odd circumstance that someone calls
          * write() in 1 byte increments */
@@ -1926,12 +1922,16 @@ gboolean
 rsvg_handle_close (RsvgHandle * handle, GError ** error)
 {
     RsvgHandlePrivate *priv;
+    gboolean result;
 
     rsvg_return_val_if_fail (handle, FALSE, error);
     priv = handle->priv;
 
-    if (priv->is_closed)
-          return TRUE;
+    if (priv->state == RSVG_HANDLE_STATE_CLOSED_OK
+        || priv->state == RSVG_HANDLE_STATE_CLOSED_ERROR) {
+        /* closing is idempotent */
+        return TRUE;
+    }
 
     if (priv->data_input_stream) {
         gboolean ret;
@@ -1943,7 +1943,15 @@ rsvg_handle_close (RsvgHandle * handle, GError ** error)
         return ret;
     }
 
-    return rsvg_handle_close_impl (handle, error);
+    result = rsvg_handle_close_impl (handle, error);
+
+    if (result) {
+        priv->state = RSVG_HANDLE_STATE_CLOSED_OK;
+    } else {
+        priv->state = RSVG_HANDLE_STATE_CLOSED_ERROR;
+    }
+
+    return result;
 }
 
 /**
@@ -1975,10 +1983,10 @@ rsvg_handle_read_stream_sync (RsvgHandle   *handle,
     xmlParserInputBufferPtr buffer;
     xmlParserInputPtr input;
     int result;
-    xmlDocPtr doc;
     GError *err = NULL;
     gboolean res = FALSE;
     const guchar *buf;
+    gssize num_read;
 
     g_return_val_if_fail (RSVG_IS_HANDLE (handle), FALSE);
     g_return_val_if_fail (G_IS_INPUT_STREAM (stream), FALSE);
@@ -1989,8 +1997,17 @@ rsvg_handle_read_stream_sync (RsvgHandle   *handle,
 
     /* detect zipped streams */
     stream = g_buffered_input_stream_new (stream);
-    if (g_buffered_input_stream_fill (G_BUFFERED_INPUT_STREAM (stream), 2, cancellable, error) != 2) {
+    num_read = g_buffered_input_stream_fill (G_BUFFERED_INPUT_STREAM (stream), 2, cancellable, error);
+    if (num_read < 2) {
         g_object_unref (stream);
+        priv->state = RSVG_HANDLE_STATE_CLOSED_ERROR;
+        if (num_read < 0) {
+            g_assert (error == NULL || *error != NULL);
+        } else {
+            g_set_error (error, rsvg_error_quark (), RSVG_ERROR_FAILED,
+                         _("Input file is too short"));
+        }
+
         return FALSE;
     }
     buf = g_buffered_input_stream_peek_buffer (G_BUFFERED_INPUT_STREAM (stream), NULL);
@@ -2034,23 +2051,23 @@ rsvg_handle_read_stream_sync (RsvgHandle   *handle,
         goto out;
     }
 
-    doc = priv->ctxt->myDoc;
-    xmlFreeParserCtxt (priv->ctxt);
-    priv->ctxt = NULL;
-
-    xmlFreeDoc (doc);
-
-    priv->finished = TRUE;
-
     res = TRUE;
 
   out:
 
+    priv->ctxt = rsvg_free_xml_parser_and_doc (priv->ctxt);
+
     g_object_unref (stream);
 
     priv->error = NULL;
     g_clear_object (&priv->cancellable);
 
+    if (res) {
+        priv->state = RSVG_HANDLE_STATE_CLOSED_OK;
+    } else {
+        priv->state = RSVG_HANDLE_STATE_CLOSED_ERROR;
+    }
+
     return res;
 }
 
@@ -2279,7 +2296,7 @@ rsvg_drawing_ctx_acquire_node_of_type (RsvgDrawingCtx * ctx, const char *url, Rs
     RsvgNode *node;
 
     node = rsvg_drawing_ctx_acquire_node (ctx, url);
-    if (node == NULL || rsvg_node_type (node) != type) {
+    if (node == NULL || rsvg_node_get_type (node) != type) {
         rsvg_drawing_ctx_release_node (ctx, node);
         return NULL;
     }
@@ -2309,6 +2326,61 @@ rsvg_drawing_ctx_release_node (RsvgDrawingCtx * ctx, RsvgNode *node)
   ctx->acquired_nodes = g_slist_remove (ctx->acquired_nodes, node);
 }
 
+void
+rsvg_drawing_ctx_add_node_and_ancestors_to_stack (RsvgDrawingCtx *draw_ctx, RsvgNode *node)
+{
+    if (node) {
+        node = rsvg_node_ref (node);
+
+        while (node != NULL) {
+            draw_ctx->drawsub_stack = g_slist_prepend (draw_ctx->drawsub_stack, node);
+            node = rsvg_node_get_parent (node);
+        }
+    }
+}
+
+void
+rsvg_drawing_ctx_draw_node_from_stack (RsvgDrawingCtx *ctx, RsvgNode *node, int dominate)
+{
+    RsvgState *state;
+    GSList *stacksave;
+
+    stacksave = ctx->drawsub_stack;
+    if (stacksave) {
+        RsvgNode *stack_node = stacksave->data;
+
+        if (!rsvg_node_is_same (stack_node, node))
+            return;
+
+        ctx->drawsub_stack = stacksave->next;
+    }
+
+    state = rsvg_node_get_state (node);
+
+    if (state->visible) {
+        rsvg_state_push (ctx);
+
+        rsvg_node_draw (node, ctx, dominate);
+
+        rsvg_state_pop (ctx);
+    }
+
+    ctx->drawsub_stack = stacksave;
+}
+
+cairo_matrix_t
+rsvg_drawing_ctx_get_current_state_affine (RsvgDrawingCtx *ctx)
+{
+    return rsvg_current_state (ctx)->affine;
+}
+
+void
+rsvg_drawing_ctx_set_current_state_affine (RsvgDrawingCtx *ctx, cairo_matrix_t *affine)
+{
+    rsvg_current_state (ctx)->personal_affine =
+        rsvg_current_state (ctx)->affine = *affine;
+}
+
 void
 rsvg_render_path_builder (RsvgDrawingCtx * ctx, RsvgPathBuilder *builder)
 {
@@ -2358,7 +2430,7 @@ rsvg_get_end_marker (RsvgDrawingCtx *ctx)
 }
 
 void
-rsvg_add_clipping_rect (RsvgDrawingCtx * ctx, double x, double y, double w, double h)
+rsvg_drawing_ctx_add_clipping_rect (RsvgDrawingCtx * ctx, double x, double y, double w, double h)
 {
     ctx->render->add_clipping_rect (ctx, x, y, w, h);
 }
@@ -2378,7 +2450,7 @@ rsvg_render_free (RsvgRender * render)
 void
 rsvg_drawing_ctx_push_view_box (RsvgDrawingCtx * ctx, double w, double h)
 {
-    RsvgViewBox *vb = g_new (RsvgViewBox, 1);
+    RsvgViewBox *vb = g_new0 (RsvgViewBox, 1);
     *vb = ctx->vb;
     ctx->vb_stack = g_slist_prepend (ctx->vb_stack, vb);
     ctx->vb.rect.width = w;
@@ -2413,6 +2485,32 @@ rsvg_drawing_ctx_get_dpi (RsvgDrawingCtx *ctx, double *out_dpi_x, double *out_dp
         *out_dpi_y = ctx->dpi_y;
 }
 
+char *
+rsvg_get_url_string (const char *str, const char **out_rest)
+{
+    if (!strncmp (str, "url(", 4)) {
+        const char *p = str + 4;
+        int ix;
+
+        while (g_ascii_isspace (*p))
+            p++;
+
+        for (ix = 0; p[ix]; ix++) {
+            if (p[ix] == ')') {
+                if (out_rest)
+                    *out_rest = p + ix + 1;
+
+                return g_strndup (p, ix);
+            }
+        }
+    }
+
+    if (out_rest)
+        *out_rest = NULL;
+
+    return NULL;
+}
+
 void
 rsvg_return_if_fail_warning (const char *pretty_function, const char *expression, GError ** error)
 {
@@ -2580,3 +2678,23 @@ _rsvg_handle_acquire_stream (RsvgHandle *handle,
     g_free (uri);
     return stream;
 }
+
+/* Frees the ctxt and its ctxt->myDoc - libxml2 doesn't free them together
+ * http://xmlsoft.org/html/libxml-parser.html#xmlFreeParserCtxt
+ *
+ * Returns NULL.
+ */
+xmlParserCtxtPtr
+rsvg_free_xml_parser_and_doc (xmlParserCtxtPtr ctxt)
+{
+    if (ctxt) {
+        if (ctxt->myDoc) {
+            xmlFreeDoc (ctxt->myDoc);
+            ctxt->myDoc = NULL;
+        }
+
+        xmlFreeParserCtxt (ctxt);
+    }
+
+    return NULL;
+}
