@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set sw=4 sts=4 ts=4 expandtab: */
+/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set sw=4 sts=4 expandtab: */
 /*
    rsvg-shapes.c: Draw shapes with cairo
 
@@ -27,6 +27,8 @@
             Carl Worth <cworth@cworth.org>
 */
 
+#include "config.h"
+
 #include "rsvg-cairo-draw.h"
 #include "rsvg-cairo-render.h"
 #include "rsvg-cairo-clip.h"
@@ -40,112 +42,7 @@
 #include <string.h>
 
 #include <pango/pangocairo.h>
-
-static gboolean
-add_color_stop_to_gradient (RsvgNode *node, gpointer data)
-{
-    Gradient *gradient = data;
-    RsvgGradientStop *stop;
-
-    if (rsvg_node_type (node) != RSVG_NODE_TYPE_STOP)
-        return TRUE; /* just ignore this node */
-
-    stop = (RsvgGradientStop *) node;
-
-    if (!stop->is_valid) {
-        /* Don't add any more stops. */
-        return FALSE;
-    }
-
-    gradient_add_color_stop (gradient, stop->offset, stop->rgba);
-
-    return TRUE;
-}
-
-static void
-add_color_stops_to_gradient (Gradient *gradient, RsvgNode *node)
-{
-    rsvg_node_foreach_child (node, add_color_stop_to_gradient, gradient);
-}
-
-static Gradient *
-linear_gradient_to_rust (RsvgLinearGradient *linear)
-{
-    Gradient *gradient;
-
-    gradient = gradient_linear_new (linear->hasx1 ? &linear->x1 : NULL,
-                                    linear->hasy1 ? &linear->y1 : NULL,
-                                    linear->hasx2 ? &linear->x2 : NULL,
-                                    linear->hasy2 ? &linear->y2 : NULL,
-                                    linear->hasbbox ? &linear->obj_bbox : NULL,
-                                    linear->hastransform ? &linear->affine : NULL,
-                                    linear->hasspread ? &linear->spread : NULL,
-                                    linear->fallback);
-
-    add_color_stops_to_gradient (gradient, (RsvgNode *) linear);
-
-    return gradient;
-}
-
-static Gradient *
-radial_gradient_to_rust (RsvgRadialGradient *radial)
-{
-    Gradient *gradient;
-
-    gradient = gradient_radial_new (radial->hascx ? &radial->cx : NULL,
-                                    radial->hascy ? &radial->cy : NULL,
-                                    radial->hasr  ? &radial->r  : NULL,
-                                    radial->hasfx ? &radial->fx : NULL,
-                                    radial->hasfy ? &radial->fy : NULL,
-                                    radial->hasbbox ? &radial->obj_bbox : NULL,
-                                    radial->hastransform ? &radial->affine : NULL,
-                                    radial->hasspread ? &radial->spread : NULL,
-                                    radial->fallback);
-
-    add_color_stops_to_gradient (gradient, (RsvgNode *) radial);
-
-    return gradient;
-}
-
-Gradient *
-rsvg_gradient_node_to_rust_gradient (RsvgNode *node)
-{
-    if (rsvg_node_type (node) == RSVG_NODE_TYPE_LINEAR_GRADIENT) {
-        return linear_gradient_to_rust ((RsvgLinearGradient *) node);
-    } else if (rsvg_node_type (node) == RSVG_NODE_TYPE_RADIAL_GRADIENT) {
-        return radial_gradient_to_rust ((RsvgRadialGradient *) node);
-    } else {
-        return NULL;
-    }
-}
-
-static void
-_set_source_rsvg_linear_gradient (RsvgDrawingCtx *ctx,
-                                  RsvgLinearGradient *linear,
-                                  guint8 opacity, RsvgBbox bbox)
-{
-    Gradient *gradient;
-
-    gradient = linear_gradient_to_rust (linear);
-
-    gradient_resolve_fallbacks_and_set_pattern (gradient, ctx, opacity, bbox);
-
-    gradient_destroy (gradient);
-}
-
-static void
-_set_source_rsvg_radial_gradient (RsvgDrawingCtx * ctx,
-                                  RsvgRadialGradient * radial,
-                                  guint8 opacity, RsvgBbox bbox)
-{
-    Gradient *gradient;
-
-    gradient = radial_gradient_to_rust (radial);
-
-    gradient_resolve_fallbacks_and_set_pattern (gradient, ctx, opacity, bbox);
-
-    gradient_destroy (gradient);
-}
+#include <pango/pangofc-fontmap.h>
 
 static void
 _set_source_rsvg_solid_color (RsvgDrawingCtx * ctx,
@@ -167,180 +64,57 @@ _set_source_rsvg_solid_color (RsvgDrawingCtx * ctx,
     cairo_set_source_rgba (cr, r, g, b, a);
 }
 
-static void
-_set_source_rsvg_pattern (RsvgDrawingCtx * ctx,
-                          RsvgPattern * rsvg_pattern, guint8 opacity, RsvgBbox bbox)
-{
-    RsvgCairoRender *render = RSVG_CAIRO_RENDER (ctx->render);
-    RsvgPattern local_pattern = *rsvg_pattern;
-    cairo_t *cr_render, *cr_pattern;
-    cairo_pattern_t *pattern;
-    cairo_surface_t *surface;
-    cairo_matrix_t matrix;
-    cairo_matrix_t affine, caffine, taffine;
-    double bbwscale, bbhscale, scwscale, schscale;
-    double patternw, patternh, patternx, patterny;
-    int pw, ph;
-
-    rsvg_pattern = &local_pattern;
-    rsvg_pattern_fix_fallback (ctx, rsvg_pattern);
-    cr_render = render->cr;
-
-    if (rsvg_pattern->obj_bbox)
-        rsvg_drawing_ctx_push_view_box (ctx, 1., 1.);
-
-    patternx = rsvg_length_normalize (&rsvg_pattern->x, ctx);
-    patterny = rsvg_length_normalize (&rsvg_pattern->y, ctx);
-    patternw = rsvg_length_normalize (&rsvg_pattern->width, ctx);
-    patternh = rsvg_length_normalize (&rsvg_pattern->height, ctx);
-
-    if (rsvg_pattern->obj_bbox)
-        rsvg_drawing_ctx_pop_view_box (ctx);
-
-
-    /* Work out the size of the rectangle so it takes into account the object bounding box */
-
-
-    if (rsvg_pattern->obj_bbox) {
-        bbwscale = bbox.rect.width;
-        bbhscale = bbox.rect.height;
-    } else {
-        bbwscale = 1.0;
-        bbhscale = 1.0;
-    }
-
-    cairo_matrix_multiply (&taffine, &rsvg_pattern->affine, &rsvg_current_state (ctx)->affine);
-
-    scwscale = sqrt (taffine.xx * taffine.xx + taffine.xy * taffine.xy);
-    schscale = sqrt (taffine.yx * taffine.yx + taffine.yy * taffine.yy);
-
-    pw = patternw * bbwscale * scwscale;
-    ph = patternh * bbhscale * schscale;
-
-    scwscale = (double) pw / (double) (patternw * bbwscale);
-    schscale = (double) ph / (double) (patternh * bbhscale);
-
-    surface = cairo_surface_create_similar (cairo_get_target (cr_render),
-                                            CAIRO_CONTENT_COLOR_ALPHA, pw, ph);
-    cr_pattern = cairo_create (surface);
-
-    /* Create the pattern coordinate system */
-    if (rsvg_pattern->obj_bbox) {
-        /* subtract the pattern origin */
-        cairo_matrix_init_translate (&affine,
-                                     bbox.rect.x + patternx * bbox.rect.width,
-                                     bbox.rect.y + patterny * bbox.rect.height);
-    } else {
-        /* subtract the pattern origin */
-        cairo_matrix_init_translate (&affine, patternx, patterny);
-    }
-    /* Apply the pattern transform */
-    cairo_matrix_multiply (&affine, &affine, &rsvg_pattern->affine);
-
-    /* Create the pattern contents coordinate system */
-    if (rsvg_pattern->vbox.active) {
-        /* If there is a vbox, use that */
-        double w, h, x, y;
-        w = patternw * bbwscale;
-        h = patternh * bbhscale;
-        x = 0;
-        y = 0;
-        rsvg_preserve_aspect_ratio (rsvg_pattern->preserve_aspect_ratio,
-                                    rsvg_pattern->vbox.rect.width, rsvg_pattern->vbox.rect.height,
-                                    &w, &h, &x, &y);
-
-        x -= rsvg_pattern->vbox.rect.x * w / rsvg_pattern->vbox.rect.width;
-        y -= rsvg_pattern->vbox.rect.y * h / rsvg_pattern->vbox.rect.height;
-
-        cairo_matrix_init (&caffine,
-                           w / rsvg_pattern->vbox.rect.width,
-                           0,
-                           0,
-                           h / rsvg_pattern->vbox.rect.height,
-                           x,
-                           y);
-        rsvg_drawing_ctx_push_view_box (ctx, rsvg_pattern->vbox.rect.width, rsvg_pattern->vbox.rect.height);
-    } else if (rsvg_pattern->obj_cbbox) {
-        /* If coords are in terms of the bounding box, use them */
-        cairo_matrix_init_scale (&caffine, bbox.rect.width, bbox.rect.height);
-        rsvg_drawing_ctx_push_view_box (ctx, 1., 1.);
-    } else {
-        cairo_matrix_init_identity (&caffine);
-    }
-
-    if (scwscale != 1.0 || schscale != 1.0) {
-        cairo_matrix_t scalematrix;
-
-        cairo_matrix_init_scale (&scalematrix, scwscale, schscale);
-        cairo_matrix_multiply (&caffine, &caffine, &scalematrix);
-        cairo_matrix_init_scale (&scalematrix, 1. / scwscale, 1. / schscale);
-        cairo_matrix_multiply (&affine, &scalematrix, &affine);
-    }
-
-    /* Draw to another surface */
-    render->cr = cr_pattern;
-
-    /* Set up transformations to be determined by the contents units */
-    rsvg_state_push (ctx);
-    rsvg_current_state (ctx)->personal_affine =
-            rsvg_current_state (ctx)->affine = caffine;
-
-    /* Draw everything */
-    _rsvg_node_draw_children ((RsvgNode *) rsvg_pattern, ctx, 2);
-    /* Return to the original coordinate system */
-    rsvg_state_pop (ctx);
-
-    /* Set the render to draw where it used to */
-    render->cr = cr_render;
-
-    pattern = cairo_pattern_create_for_surface (surface);
-    cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
-
-    matrix = affine;
-    if (cairo_matrix_invert (&matrix) != CAIRO_STATUS_SUCCESS)
-      goto out;
-
-    cairo_pattern_set_matrix (pattern, &matrix);
-    cairo_pattern_set_filter (pattern, CAIRO_FILTER_BEST);
-
-    cairo_set_source (cr_render, pattern);
-
-    cairo_pattern_destroy (pattern);
-    cairo_destroy (cr_pattern);
-    cairo_surface_destroy (surface);
-
-  out:
-    if (rsvg_pattern->obj_cbbox || rsvg_pattern->vbox.active)
-        rsvg_drawing_ctx_pop_view_box (ctx);
-}
-
 /* note: _set_source_rsvg_paint_server does not change cairo's CTM */
-static void
+static gboolean
 _set_source_rsvg_paint_server (RsvgDrawingCtx * ctx,
                                guint32 current_color_rgb,
                                RsvgPaintServer * ps,
                                guint8 opacity, RsvgBbox bbox, guint32 current_color)
 {
     RsvgNode *node;
+    gboolean had_paint_server;
+    gboolean use_alternate;
+
+    had_paint_server = FALSE;
 
     switch (ps->type) {
     case RSVG_PAINT_SERVER_IRI:
-        node = rsvg_drawing_ctx_acquire_node (ctx, ps->core.iri);
-        if (node == NULL)
-            break;
-        else if (rsvg_node_type (node) == RSVG_NODE_TYPE_LINEAR_GRADIENT)
-            _set_source_rsvg_linear_gradient (ctx, (RsvgLinearGradient *) node, opacity, bbox);
-        else if (rsvg_node_type (node) == RSVG_NODE_TYPE_RADIAL_GRADIENT)
-            _set_source_rsvg_radial_gradient (ctx, (RsvgRadialGradient *) node, opacity, bbox);
-        else if (rsvg_node_type (node) == RSVG_NODE_TYPE_PATTERN)
-            _set_source_rsvg_pattern (ctx, (RsvgPattern *) node, opacity, bbox);
+        use_alternate = FALSE;
+
+        node = rsvg_drawing_ctx_acquire_node (ctx, ps->core.iri->iri_str);
+        if (node == NULL) {
+            use_alternate = TRUE;
+        } else if (rsvg_node_get_type (node) == RSVG_NODE_TYPE_LINEAR_GRADIENT
+                   || rsvg_node_get_type (node) == RSVG_NODE_TYPE_RADIAL_GRADIENT) {
+            if (gradient_resolve_fallbacks_and_set_pattern (node, ctx, opacity, bbox)) {
+                had_paint_server = TRUE;
+            } else {
+                use_alternate = TRUE;
+            }
+        } else if (rsvg_node_get_type (node) == RSVG_NODE_TYPE_PATTERN) {
+            if (pattern_resolve_fallbacks_and_set_pattern (node, ctx, bbox)) {
+                had_paint_server = TRUE;
+            } else {
+                use_alternate = TRUE;
+            }
+        }
+
+        if (use_alternate) {
+            if (ps->core.iri->has_alternate) {
+                _set_source_rsvg_solid_color (ctx, &ps->core.iri->alternate, opacity, current_color);
+                had_paint_server = TRUE;
+            }
+        }
 
         rsvg_drawing_ctx_release_node (ctx, node);
         break;
     case RSVG_PAINT_SERVER_SOLID:
         _set_source_rsvg_solid_color (ctx, ps->core.color, opacity, current_color);
+        had_paint_server = TRUE;
         break;
     }
+
+    return had_paint_server;
 }
 
 static void
@@ -358,6 +132,7 @@ _set_rsvg_affine (RsvgCairoRender * render, cairo_matrix_t *affine)
     cairo_set_matrix (cr, &matrix);
 }
 
+#ifdef HAVE_PANGOFT2
 static cairo_font_options_t *
 get_font_options_for_testing (void)
 {
@@ -371,6 +146,46 @@ get_font_options_for_testing (void)
     return options;
 }
 
+static void
+set_font_options_for_testing (PangoContext *context)
+{
+    cairo_font_options_t *font_options;
+
+    font_options = get_font_options_for_testing ();
+    pango_cairo_context_set_font_options (context, font_options);
+    cairo_font_options_destroy (font_options);
+}
+
+static void
+create_font_config_for_testing (RsvgCairoRender *render)
+{
+    const char *font_path = SRCDIR "/tests/resources/LiberationSans-Regular.ttf";
+
+    if (render->font_config_for_testing != NULL)
+        return;
+
+    render->font_config_for_testing = FcConfigCreate ();
+
+    if (!FcConfigAppFontAddFile (render->font_config_for_testing, (const FcChar8 *) font_path)) {
+        g_error ("Could not load font file \"%s\" for tests; aborting", font_path);
+    }
+}
+
+static PangoFontMap *
+get_font_map_for_testing (RsvgCairoRender *render)
+{
+    create_font_config_for_testing (render);
+
+    if (!render->font_map_for_testing) {
+        render->font_map_for_testing = pango_cairo_font_map_new_for_font_type (CAIRO_FONT_TYPE_FT);
+        pango_fc_font_map_set_config (PANGO_FC_FONT_MAP (render->font_map_for_testing),
+                                      render->font_config_for_testing);
+    }
+
+    return render->font_map_for_testing;
+}
+#endif
+
 PangoContext *
 rsvg_cairo_create_pango_context (RsvgDrawingCtx * ctx)
 {
@@ -379,19 +194,27 @@ rsvg_cairo_create_pango_context (RsvgDrawingCtx * ctx)
     RsvgCairoRender *render = RSVG_CAIRO_RENDER (ctx->render);
     double dpi_y;
 
-    fontmap = pango_cairo_font_map_get_default ();
+#ifdef HAVE_PANGOFT2
+    if (ctx->is_testing) {
+        fontmap = get_font_map_for_testing (render);
+    } else {
+#endif
+        fontmap = pango_cairo_font_map_get_default ();
+#ifdef HAVE_PANGOFT2
+    }
+#endif
+
     context = pango_font_map_create_context (fontmap);
     pango_cairo_update_context (render->cr, context);
 
     rsvg_drawing_ctx_get_dpi (ctx, NULL, &dpi_y);
     pango_cairo_context_set_resolution (context, dpi_y);
 
+#ifdef HAVE_PANGOFT2
     if (ctx->is_testing) {
-        cairo_font_options_t *font_options;
-        font_options = get_font_options_for_testing ();
-        pango_cairo_context_set_font_options (context, font_options);
-        cairo_font_options_destroy (font_options);
+        set_font_options_for_testing (context);
     }
+#endif
 
     return context;
 }
@@ -417,12 +240,16 @@ rsvg_cairo_render_pango_layout (RsvgDrawingCtx * ctx, PangoLayout * layout, doub
     PangoGravity gravity = pango_context_get_gravity (pango_layout_get_context (layout));
     double rotation;
 
+    pango_layout_get_extents(layout, &ink, NULL);
+
+    if (ink.width == 0 || ink.height == 0) {
+        return;
+    }
+
     cairo_set_antialias (render->cr, state->text_rendering_type);
 
     _set_rsvg_affine (render, &state->affine);
 
-    pango_layout_get_extents (layout, &ink, NULL);
-
     rsvg_bbox_init (&bbox, &state->affine);
     if (PANGO_GRAVITY_IS_VERTICAL (gravity)) {
         bbox.rect.x = x + (ink.x - ink.height) / (double)PANGO_SCALE;
@@ -442,14 +269,19 @@ rsvg_cairo_render_pango_layout (RsvgDrawingCtx * ctx, PangoLayout * layout, doub
         cairo_save (render->cr);
         cairo_move_to (render->cr, x, y);
         rsvg_bbox_insert (&render->bbox, &bbox);
-        _set_source_rsvg_paint_server (ctx,
-                                       state->current_color,
-                                       state->fill,
-                                       state->fill_opacity,
-                                       bbox, rsvg_current_state (ctx)->current_color);
-        if (rotation != 0.)
-            cairo_rotate (render->cr, -rotation);
-        pango_cairo_show_layout (render->cr, layout);
+
+        if (_set_source_rsvg_paint_server (ctx,
+                                           state->current_color,
+                                           state->fill,
+                                           state->fill_opacity,
+                                           bbox, rsvg_current_state (ctx)->current_color)) {
+            if (rotation != 0.)
+                cairo_rotate (render->cr, -rotation);
+
+            pango_cairo_update_layout (render->cr, layout);
+            pango_cairo_show_layout (render->cr, layout);
+        }
+
         cairo_restore (render->cr);
     }
 
@@ -458,19 +290,22 @@ rsvg_cairo_render_pango_layout (RsvgDrawingCtx * ctx, PangoLayout * layout, doub
         cairo_move_to (render->cr, x, y);
         rsvg_bbox_insert (&render->bbox, &bbox);
 
-        _set_source_rsvg_paint_server (ctx,
-                                       state->current_color,
-                                       state->stroke,
-                                       state->stroke_opacity,
-                                       bbox, rsvg_current_state (ctx)->current_color);
+        if (_set_source_rsvg_paint_server (ctx,
+                                           state->current_color,
+                                           state->stroke,
+                                           state->stroke_opacity,
+                                           bbox, rsvg_current_state (ctx)->current_color)) {
+            if (rotation != 0.)
+                cairo_rotate (render->cr, -rotation);
 
-        if (rotation != 0.)
-            cairo_rotate (render->cr, -rotation);
-        pango_cairo_layout_path (render->cr, layout);
+            pango_cairo_update_layout (render->cr, layout);
+            pango_cairo_layout_path (render->cr, layout);
 
-        setup_cr_for_stroke (render->cr, ctx, state);
+            setup_cr_for_stroke (render->cr, ctx, state);
+
+            cairo_stroke (render->cr);
+        }
 
-        cairo_stroke (render->cr);
         cairo_restore (render->cr);
     }
 }
@@ -506,6 +341,19 @@ rsvg_cairo_render_path_builder (RsvgDrawingCtx * ctx, RsvgPathBuilder *builder)
        _rendering_ time speedups, are these rather expensive operations
        really needed here? */
 
+    /* FIXME: See https://www.w3.org/TR/SVG/coords.html#ObjectBoundingBox for
+     * discussion on how to compute bounding boxes to be used for viewports and
+     * clipping.  It looks like we should be using cairo_path_extents() for
+     * that, not cairo_fill_extents().
+     *
+     * We may need to maintain *two* sets of bounding boxes - one for
+     * viewports/clipping, and one for user applications like a
+     * rsvg_compute_ink_rect() function in the future.
+     *
+     * See https://bugzilla.gnome.org/show_bug.cgi?id=760112 for discussion of a
+     * public API to get the ink rectangle.
+     */
+
     /* Bounding box for fill
      *
      * Unlike the case for stroke, for fills we always compute the bounding box.
@@ -546,27 +394,27 @@ rsvg_cairo_render_path_builder (RsvgDrawingCtx * ctx, RsvgPathBuilder *builder)
 
         opacity = state->fill_opacity;
 
-        _set_source_rsvg_paint_server (ctx,
-                                       state->current_color,
-                                       state->fill,
-                                       opacity, bbox, rsvg_current_state (ctx)->current_color);
-
-        if (state->stroke != NULL)
-            cairo_fill_preserve (cr);
-        else
-            cairo_fill (cr);
+        if (_set_source_rsvg_paint_server (ctx,
+                                           state->current_color,
+                                           state->fill,
+                                           opacity, bbox, rsvg_current_state (ctx)->current_color)) {
+            if (state->stroke != NULL)
+                cairo_fill_preserve (cr);
+            else
+                cairo_fill (cr);
+        }
     }
 
     if (state->stroke != NULL) {
         int opacity;
         opacity = state->stroke_opacity;
 
-        _set_source_rsvg_paint_server (ctx,
-                                       state->current_color,
-                                       state->stroke,
-                                       opacity, bbox, rsvg_current_state (ctx)->current_color);
-
-        cairo_stroke (cr);
+        if (_set_source_rsvg_paint_server (ctx,
+                                           state->current_color,
+                                           state->stroke,
+                                           opacity, bbox, rsvg_current_state (ctx)->current_color)) {
+            cairo_stroke (cr);
+        }
     }
 
     cairo_new_path (cr); /* clear the path in case stroke == fill == NULL; otherwise we leave it around from computing the bounding box */
@@ -644,8 +492,34 @@ rsvg_cairo_get_cairo_context (RsvgDrawingCtx *ctx)
     return render->cr;
 }
 
+/* FIXME: Usage of this function is more less a hack.  Some code does this:
+ *
+ *   save_cr = rsvg_cairo_get_cairo_context (ctx);
+ *
+ *   some_surface = create_surface ();
+ *
+ *   cr = cairo_create (some_surface);
+ *
+ *   rsvg_cairo_set_cairo_context (ctx, cr);
+ *
+ *   ... draw with ctx but to that temporary surface
+ *
+ *   rsvg_cairo_set_cairo_context (ctx, save_cr);
+ *
+ * It would be better to have an explicit push/pop for the cairo_t, or
+ * pushing a temporary surface, or something that does not involve
+ * monkeypatching the cr directly.
+ */
+void
+rsvg_cairo_set_cairo_context (RsvgDrawingCtx *ctx, cairo_t *cr)
+{
+    RsvgCairoRender *render = RSVG_CAIRO_RENDER (ctx->render);
+
+    render->cr = cr;
+}
+
 static void
-rsvg_cairo_generate_mask (cairo_t * cr, RsvgMask * self, RsvgDrawingCtx * ctx, RsvgBbox * bbox)
+rsvg_cairo_generate_mask (cairo_t * cr, RsvgNode *node_mask, RsvgDrawingCtx *ctx, RsvgBbox *bbox)
 {
     RsvgCairoRender *render = RSVG_CAIRO_RENDER (ctx->render);
     cairo_surface_t *surface;
@@ -657,6 +531,10 @@ rsvg_cairo_generate_mask (cairo_t * cr, RsvgMask * self, RsvgDrawingCtx * ctx, R
     cairo_matrix_t affinesave;
     double sx, sy, sw, sh;
     gboolean nest = cr != render->initial_cr;
+    RsvgMask *self;
+
+    g_assert (rsvg_node_get_type (node_mask) == RSVG_NODE_TYPE_MASK);
+    self = rsvg_rust_cnode_get_impl (node_mask);
 
     surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, width, height);
     if (cairo_surface_status (surface) != CAIRO_STATUS_SUCCESS) {
@@ -694,6 +572,8 @@ rsvg_cairo_generate_mask (cairo_t * cr, RsvgMask * self, RsvgDrawingCtx * ctx, R
     /* Horribly dirty hack to have the bbox premultiplied to everything */
     if (self->contentunits == objectBoundingBox) {
         cairo_matrix_t bbtransform;
+        RsvgState *mask_state;
+
         cairo_matrix_init (&bbtransform,
                            bbox->rect.width,
                            0,
@@ -701,18 +581,25 @@ rsvg_cairo_generate_mask (cairo_t * cr, RsvgMask * self, RsvgDrawingCtx * ctx, R
                            bbox->rect.height,
                            bbox->rect.x,
                            bbox->rect.y);
-        affinesave = self->super.state->affine;
-        cairo_matrix_multiply (&self->super.state->affine, &bbtransform, &self->super.state->affine);
+
+        mask_state = rsvg_node_get_state (node_mask);
+
+        affinesave = mask_state->affine;
+        cairo_matrix_multiply (&mask_state->affine, &bbtransform, &mask_state->affine);
         rsvg_drawing_ctx_push_view_box (ctx, 1, 1);
     }
 
     rsvg_state_push (ctx);
-    _rsvg_node_draw_children (&self->super, ctx, 0);
+    rsvg_node_draw_children (node_mask, ctx, 0);
     rsvg_state_pop (ctx);
 
     if (self->contentunits == objectBoundingBox) {
+        RsvgState *mask_state;
+
         rsvg_drawing_ctx_pop_view_box (ctx);
-        self->super.state->affine = affinesave;
+
+        mask_state = rsvg_node_get_state (node_mask);
+        mask_state->affine = affinesave;
     }
 
     render->cr = save_cr;
@@ -769,11 +656,11 @@ rsvg_cairo_push_render_stack (RsvgDrawingCtx * ctx)
         RsvgNode *node;
         node = rsvg_drawing_ctx_acquire_node_of_type (ctx, rsvg_current_state (ctx)->clip_path, RSVG_NODE_TYPE_CLIP_PATH);
         if (node) {
-            RsvgClipPath *clip_path = (RsvgClipPath *) node;
+            RsvgClipPath *clip_path = rsvg_rust_cnode_get_impl (node);
 
             switch (clip_path->units) {
             case userSpaceOnUse:
-                rsvg_cairo_clip (ctx, clip_path, NULL);
+                rsvg_cairo_clip (ctx, node, NULL);
                 break;
             case objectBoundingBox:
                 lateclip = TRUE;
@@ -818,7 +705,7 @@ rsvg_cairo_push_render_stack (RsvgDrawingCtx * ctx)
     render->cr_stack = g_list_prepend (render->cr_stack, render->cr);
     render->cr = child_cr;
 
-    bbox = g_new (RsvgBbox, 1);
+    bbox = g_new0 (RsvgBbox, 1);
     *bbox = render->bbox;
     render->bb_stack = g_list_prepend (render->bb_stack, bbox);
     rsvg_bbox_init (&render->bbox, &state->affine);
@@ -838,7 +725,7 @@ rsvg_cairo_pop_render_stack (RsvgDrawingCtx * ctx)
 {
     RsvgCairoRender *render = RSVG_CAIRO_RENDER (ctx->render);
     cairo_t *child_cr = render->cr;
-    RsvgClipPath *lateclip = NULL;
+    RsvgNode *lateclip = NULL;
     cairo_surface_t *surface = NULL;
     RsvgState *state = rsvg_current_state (ctx);
     gboolean nest, needs_destroy = FALSE;
@@ -846,10 +733,15 @@ rsvg_cairo_pop_render_stack (RsvgDrawingCtx * ctx)
     if (rsvg_current_state (ctx)->clip_path) {
         RsvgNode *node;
         node = rsvg_drawing_ctx_acquire_node_of_type (ctx, rsvg_current_state (ctx)->clip_path, RSVG_NODE_TYPE_CLIP_PATH);
-        if (node && ((RsvgClipPath *) node)->units == objectBoundingBox)
-            lateclip = (RsvgClipPath *) node;
-        else
-            rsvg_drawing_ctx_release_node (ctx, node);
+        if (node) {
+            RsvgClipPath *cp = rsvg_rust_cnode_get_impl (node);
+
+            if (cp->units == objectBoundingBox) {
+                lateclip = node;
+            } else {
+                rsvg_drawing_ctx_release_node (ctx, node);
+            }
+        }
     }
 
     if (state->opacity == 0xFF
@@ -869,7 +761,7 @@ rsvg_cairo_pop_render_stack (RsvgDrawingCtx * ctx)
         filter = rsvg_drawing_ctx_acquire_node_of_type (ctx, state->filter, RSVG_NODE_TYPE_FILTER);
         if (filter) {
             needs_destroy = TRUE;
-            surface = rsvg_filter_render ((RsvgFilter *) filter, output, ctx, &render->bbox, "2103");
+            surface = rsvg_filter_render (filter, output, ctx, &render->bbox, "2103");
             rsvg_drawing_ctx_release_node (ctx, filter);
 
             /* Don't destroy the output surface, it's owned by child_cr */
@@ -887,7 +779,7 @@ rsvg_cairo_pop_render_stack (RsvgDrawingCtx * ctx)
 
     if (lateclip) {
         rsvg_cairo_clip (ctx, lateclip, &render->bbox);
-        rsvg_drawing_ctx_release_node (ctx, (RsvgNode *) lateclip);
+        rsvg_drawing_ctx_release_node (ctx, lateclip);
     }
 
     cairo_set_operator (render->cr, state->comp_op);
@@ -897,7 +789,7 @@ rsvg_cairo_pop_render_stack (RsvgDrawingCtx * ctx)
 
         mask = rsvg_drawing_ctx_acquire_node_of_type (ctx, state->mask, RSVG_NODE_TYPE_MASK);
         if (mask) {
-            rsvg_cairo_generate_mask (render->cr, (RsvgMask *) mask, ctx, &render->bbox);
+            rsvg_cairo_generate_mask (render->cr, mask, ctx, &render->bbox);
             rsvg_drawing_ctx_release_node (ctx, mask);
         }
     } else if (state->opacity != 0xFF)
@@ -963,7 +855,7 @@ rsvg_cairo_get_surface_of_node (RsvgDrawingCtx *ctx,
     render = rsvg_cairo_render_new (cr, width, height);
     ctx->render = (RsvgRender *) render;
 
-    rsvg_node_draw (drawable, ctx, 0);
+    rsvg_drawing_ctx_draw_node_from_stack (ctx, drawable, 0);
 
     cairo_destroy (cr);
 
