@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set sw=4 sts=4 ts=4 expandtab: */
+/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set sw=4 sts=4 expandtab: */
 /*
    rsvg-shapes.c: Draw shapes with libart
 
@@ -46,6 +46,8 @@ void         rsvg_cairo_render_surface          (RsvgDrawingCtx *ctx, cairo_surf
                                                  double x, double y, double w, double h);
 G_GNUC_INTERNAL
 cairo_t *rsvg_cairo_get_cairo_context (RsvgDrawingCtx *ctx);
+G_GNUC_INTERNAL
+void rsvg_cairo_set_cairo_context (RsvgDrawingCtx *ctx, cairo_t *cr);
 
 G_GNUC_INTERNAL
 void         rsvg_cairo_push_discrete_layer	    (RsvgDrawingCtx *ctx);
