@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set sw=4 sts=4 ts=4 expandtab: */
+/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set sw=4 sts=4 expandtab: */
 /*
    rsvg-private.h: Internals of RSVG
 
@@ -123,11 +123,20 @@ struct RsvgSaxHandler {
     void (*characters) (RsvgSaxHandler * self, const char *ch, int len);
 };
 
+/* Reading state for an RsvgHandle */
+typedef enum {
+    RSVG_HANDLE_STATE_START,
+    RSVG_HANDLE_STATE_READING,
+    RSVG_HANDLE_STATE_CLOSED_OK,
+    RSVG_HANDLE_STATE_CLOSED_ERROR
+} RsvgHandleState;
+
 struct RsvgHandlePrivate {
     RsvgHandleFlags flags;
 
+    RsvgHandleState state;
+
     gboolean is_disposed;
-    gboolean is_closed;
 
     RsvgSizeFunc size_func;
     gpointer user_data;
@@ -170,16 +179,14 @@ struct RsvgHandlePrivate {
     gchar *base_uri;
     GFile *base_gfile;
 
-    gboolean finished;
-
     gboolean in_loop;		/* see get_dimension() */
 
-    gboolean first_write;
     GInputStream *data_input_stream; /* for rsvg_handle_write of svgz data */
 
     gboolean is_testing; /* Are we being run from the test suite? */
 };
 
+/* Keep this in sync with rust/src/viewbox.rs::RsvgViewBox */
 typedef struct {
     cairo_rectangle_t rect;
     gboolean active;
@@ -278,6 +285,7 @@ typedef enum {
     objectBoundingBox, userSpaceOnUse
 } RsvgCoordUnits;
 
+/* Keep this in sync with rust/src/node.rs:NodeType */
 typedef enum {
     RSVG_NODE_TYPE_INVALID = 0,
 
@@ -332,40 +340,89 @@ typedef enum {
     RSVG_NODE_TYPE_FILTER_PRIMITIVE_LAST                /* just a marker; not a valid type */
 } RsvgNodeType;
 
-typedef struct {
-    void (*free) (RsvgNode * self);
-    void (*draw) (RsvgNode * self, RsvgDrawingCtx * ctx, int dominate);
-    void (*set_atts) (RsvgNode * self, RsvgHandle * ctx, RsvgPropertyBag *atts);
-} RsvgNodeVtable;
-
-struct _RsvgNode {
-    RsvgState *state;
-    RsvgNode *parent;
-    GPtrArray *children;
-    RsvgNodeType type;
-    RsvgNodeVtable *vtable;
-};
+typedef void (* CNodeSetAtts) (RsvgNode *node, gpointer impl, RsvgHandle *handle, RsvgPropertyBag *pbag);
+typedef void (* CNodeDraw) (RsvgNode *node, gpointer impl, RsvgDrawingCtx *ctx, int dominate);
+typedef void (* CNodeFree) (gpointer impl);
 
+/* Implemented in rust/src/node.rs */
+/* Call node = rsvg_node_unref (node) when you are done with the node */
 G_GNUC_INTERNAL
-RsvgNodeType rsvg_node_type (RsvgNode *node);
+RsvgNode *rsvg_rust_cnode_new (RsvgNodeType  node_type,
+                               RsvgNode     *parent,
+                               RsvgState    *state,
+                               gpointer      impl,
+                               CNodeSetAtts  set_atts_fn,
+                               CNodeDraw     draw_fn,
+                               CNodeFree     free_fn) G_GNUC_WARN_UNUSED_RESULT;
 
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+gpointer rsvg_rust_cnode_get_impl (RsvgNode *node);
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+RsvgNodeType rsvg_node_get_type (RsvgNode *node);
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+gboolean rsvg_node_is_same (RsvgNode *node1, RsvgNode *node2);
+
+/* Implemented in rust/src/node.rs */
+/* Call this as newref = rsvg_node_ref (node);  You don't own the node anymore, just the newref! */
+G_GNUC_INTERNAL
+RsvgNode *rsvg_node_ref (RsvgNode *node) G_GNUC_WARN_UNUSED_RESULT;
+
+/* Implemented in rust/src/node.rs */
+/* Call this as node = rsvg_node_unref (node);  Then node will be NULL and you don't own it anymore! */
+G_GNUC_INTERNAL
+RsvgNode *rsvg_node_unref (RsvgNode *node) G_GNUC_WARN_UNUSED_RESULT;
+
+/* Implemented in rust/src/node.rs */
 G_GNUC_INTERNAL
 RsvgState *rsvg_node_get_state (RsvgNode *node);
 
+/* Implemented in rust/src/node.rs
+ *
+ * Returns a new strong reference to the parent (or NULL); use rsvg_node_unref()
+ * when you are done.
+ */
 G_GNUC_INTERNAL
-RsvgNode *rsvg_node_get_parent (RsvgNode *node);
+RsvgNode *rsvg_node_get_parent (RsvgNode *node) G_GNUC_WARN_UNUSED_RESULT;
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+void rsvg_node_add_child (RsvgNode *node, RsvgNode *child);
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+void rsvg_node_set_atts (RsvgNode *node, RsvgHandle *handle, RsvgPropertyBag *atts);
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+void rsvg_node_draw (RsvgNode *node, RsvgDrawingCtx *draw, int dominate);
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+void rsvg_node_set_attribute_parse_error (RsvgNode *node, const char *attr_name, const char *description);
 
 /* Used to iterate among a node's children with rsvg_node_foreach_child().
  * If this caller-supplied function returns FALSE, iteration will stop.
  * Otherwise, iteration will continue to the next child node.
+ *
+ * Keep this in sync with rust/src/node.rs:NodeForeachChild
  */
 typedef gboolean (* RsvgNodeForeachChildFn) (RsvgNode *node, gpointer data);
 
+/* Implemented in rust/src/node.rs */
 G_GNUC_INTERNAL
 void rsvg_node_foreach_child (RsvgNode *node, RsvgNodeForeachChildFn fn, gpointer data);
+/* generic function for drawing all of the children of a particular node */
+
+/* Implemented in rust/src/node.rs */
+G_GNUC_INTERNAL
+void rsvg_node_draw_children (RsvgNode *node, RsvgDrawingCtx *ctx, int dominate);
 
 struct _RsvgNodeChars {
-    RsvgNode super;
     GString *contents;
 };
 
@@ -406,6 +463,11 @@ RsvgNode *rsvg_drawing_ctx_acquire_node_of_type (RsvgDrawingCtx * ctx, const cha
 G_GNUC_INTERNAL
 void rsvg_drawing_ctx_release_node              (RsvgDrawingCtx * ctx, RsvgNode *node);
 
+G_GNUC_INTERNAL
+void rsvg_drawing_ctx_add_node_and_ancestors_to_stack (RsvgDrawingCtx *draw_ctx, RsvgNode *node);
+G_GNUC_INTERNAL
+void rsvg_drawing_ctx_draw_node_from_stack            (RsvgDrawingCtx *ctx, RsvgNode *node, int dominate);
+
 G_GNUC_INTERNAL
 void rsvg_render_path_builder   (RsvgDrawingCtx * ctx, RsvgPathBuilder *builder);
 G_GNUC_INTERNAL
@@ -425,7 +487,7 @@ const char *rsvg_get_end_marker (RsvgDrawingCtx *ctx);
 G_GNUC_INTERNAL
 void rsvg_render_free           (RsvgRender * render);
 G_GNUC_INTERNAL
-void rsvg_add_clipping_rect     (RsvgDrawingCtx * ctx, double x, double y, double w, double h);
+void rsvg_drawing_ctx_add_clipping_rect     (RsvgDrawingCtx * ctx, double x, double y, double w, double h);
 G_GNUC_INTERNAL
 cairo_surface_t *rsvg_cairo_surface_from_pixbuf (const GdkPixbuf *pixbuf);
 G_GNUC_INTERNAL
@@ -462,6 +524,12 @@ double rsvg_length_hand_normalize (const RsvgLength *length,
 G_GNUC_INTERNAL
 double rsvg_drawing_ctx_get_normalized_font_size (RsvgDrawingCtx * ctx);
 
+G_GNUC_INTERNAL
+cairo_matrix_t rsvg_drawing_ctx_get_current_state_affine (RsvgDrawingCtx *ctx);
+
+G_GNUC_INTERNAL
+void rsvg_drawing_ctx_set_current_state_affine (RsvgDrawingCtx *ctx, cairo_matrix_t *affine);
+
 G_GNUC_INTERNAL
 double _rsvg_css_accumulate_baseline_shift (RsvgState * state, RsvgDrawingCtx * ctx);
 
@@ -482,7 +550,7 @@ void rsvg_drawing_ctx_get_dpi (RsvgDrawingCtx *ctx, double *out_dpi_x, double *o
 G_GNUC_INTERNAL
 void rsvg_SAX_handler_struct_init (void);
 G_GNUC_INTERNAL
-char *rsvg_get_url_string (const char *str);
+char *rsvg_get_url_string (const char *str, const char **out_rest);
 G_GNUC_INTERNAL
 void rsvg_return_if_fail_warning (const char *pretty_function,
                                   const char *expression, GError ** error);
@@ -499,6 +567,9 @@ GInputStream *_rsvg_handle_acquire_stream (RsvgHandle *handle,
                                            char **content_type,
                                            GError **error);
 
+G_GNUC_INTERNAL
+xmlParserCtxtPtr rsvg_free_xml_parser_and_doc (xmlParserCtxtPtr ctxt) G_GNUC_WARN_UNUSED_RESULT;
+
 
 #define rsvg_return_if_fail(expr, error)    G_STMT_START{			\
      if G_LIKELY(expr) { } else                                     \
