@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set sw=4 sts=4 ts=4 expandtab: */
+/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set sw=4 sts=4 expandtab: */
 /*
    rsvg-styles.c: Handle SVG styles
 
@@ -25,10 +25,10 @@
 */
 #include "config.h"
 
+#include <errno.h>
 #include <string.h>
 #include <math.h>
 
-#include "rsvg.h"
 #include "rsvg-private.h"
 #include "rsvg-filter.h"
 #include "rsvg-css.h"
@@ -60,27 +60,12 @@ typedef struct _StyleValueData {
     gboolean important;
 } StyleValueData;
 
-/*
- * _rsvg_cairo_matrix_init_shear: Set up a shearing matrix.
- * @dst: Where to store the resulting affine transform.
- * @theta: Shear angle in degrees.
- *
- * Sets up a shearing matrix. In the standard libart coordinate system
- * and a small value for theta, || becomes \\. Horizontal lines remain
- * unchanged.
- **/
-static void
-_rsvg_cairo_matrix_init_shear (cairo_matrix_t *dst, double theta)
-{
-  cairo_matrix_init (dst, 1., 0., tan (theta * M_PI / 180.0), 1., 0., 0);
-}
-
 static StyleValueData *
 style_value_data_new (const gchar *value, gboolean important)
 {
     StyleValueData *ret;
 
-    ret = g_new (StyleValueData, 1);
+    ret = g_new0 (StyleValueData, 1);
     ret->value = g_strdup (value);
     ret->important = important;
 
@@ -118,10 +103,17 @@ rsvg_state_init (RsvgState * state)
     state->miter_limit = 4;
     state->cap = CAIRO_LINE_CAP_BUTT;
     state->join = CAIRO_LINE_JOIN_MITER;
-    state->stop_color = 0x00;
-    state->stop_color_mode = STOP_COLOR_UNSPECIFIED;
-    state->stop_opacity = 0xff;
-    state->stop_opacity_mode = STOP_OPACITY_UNSPECIFIED;
+
+    /* The following two start as INHERIT, even though has_stop_color and
+     * has_stop_opacity get initialized to FALSE below.  This is so that the
+     * first pass of rsvg_state_inherit_run(), called from
+     * rsvg_state_reconstruct() from the "stop" element code, will correctly
+     * initialize the destination state from the toplevel element.
+     *
+     */
+    state->stop_color.kind = RSVG_CSS_COLOR_SPEC_INHERIT;
+    state->stop_opacity.kind = RSVG_OPACITY_INHERIT;
+
     state->fill_rule = CAIRO_FILL_RULE_WINDING;
     state->clip_rule = CAIRO_FILL_RULE_WINDING;
     state->enable_background = RSVG_ENABLE_BACKGROUND_ACCUMULATE;
@@ -296,7 +288,7 @@ rsvg_state_clone (RsvgState * dst, const RsvgState * src)
     dst->styles = g_hash_table_ref (src->styles);
 
     if (src->dash.n_dash > 0) {
-        dst->dash.dash = g_new (gdouble, src->dash.n_dash);
+        dst->dash.dash = g_new0 (gdouble, src->dash.n_dash);
         for (i = 0; i < src->dash.n_dash; i++)
             dst->dash.dash[i] = src->dash.dash[i];
     }
@@ -354,12 +346,16 @@ rsvg_state_inherit_run (RsvgState * dst, const RsvgState * src,
     if (function (dst->has_join, src->has_join))
         dst->join = src->join;
     if (function (dst->has_stop_color, src->has_stop_color)) {
-        dst->stop_color = src->stop_color;
-        dst->stop_color_mode = src->stop_color_mode;
+        if (dst->stop_color.kind == RSVG_CSS_COLOR_SPEC_INHERIT) {
+            dst->has_stop_color = TRUE;
+            dst->stop_color = src->stop_color;
+        }
     }
     if (function (dst->has_stop_opacity, src->has_stop_opacity)) {
-        dst->stop_opacity = src->stop_opacity;
-        dst->stop_opacity_mode = src->stop_opacity_mode;
+        if (dst->stop_opacity.kind == RSVG_OPACITY_INHERIT) {
+            dst->has_stop_opacity = TRUE;
+            dst->stop_opacity = src->stop_opacity;
+        }
     }
     if (function (dst->has_cond, src->has_cond))
         dst->cond_true = src->cond_true;
@@ -423,7 +419,7 @@ rsvg_state_inherit_run (RsvgState * dst, const RsvgState * src,
         if (dst->has_dash)
             g_free (dst->dash.dash);
 
-        dst->dash.dash = g_new (gdouble, src->dash.n_dash);
+        dst->dash.dash = g_new0 (gdouble, src->dash.n_dash);
         dst->dash.n_dash = src->dash.n_dash;
         for (i = 0; i < src->dash.n_dash; i++)
             dst->dash.dash[i] = src->dash.dash[i];
@@ -524,8 +520,7 @@ rsvg_state_inherit (RsvgState * dst, const RsvgState * src)
 
 /* Parse a CSS2 style argument, setting the SVG context attributes. */
 static void
-rsvg_parse_style_pair (RsvgHandle * ctx,
-                       RsvgState * state,
+rsvg_parse_style_pair (RsvgState * state,
                        const gchar * name,
                        const gchar * value,
                        gboolean important)
@@ -543,21 +538,85 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
                          (gpointer) g_strdup (name),
                          (gpointer) style_value_data_new (value, important));
 
-    if (g_str_equal (name, "color"))
-        state->current_color = rsvg_css_parse_color (value, &state->has_current_color);
-    else if (g_str_equal (name, "opacity"))
-        state->opacity = rsvg_css_parse_opacity (value);
-    else if (g_str_equal (name, "flood-color"))
-        state->flood_color = rsvg_css_parse_color (value, &state->has_flood_color);
-    else if (g_str_equal (name, "flood-opacity")) {
-        state->flood_opacity = rsvg_css_parse_opacity (value);
+    if (g_str_equal (name, "color")) {
+        RsvgCssColorSpec spec;
+
+        spec = rsvg_css_parse_color (value, ALLOW_INHERIT_YES, ALLOW_CURRENT_COLOR_NO);
+        switch (spec.kind) {
+        case RSVG_CSS_COLOR_SPEC_INHERIT:
+            /* FIXME: we should inherit; see how stop-color is handled in rsvg-styles.c */
+            state->has_current_color = FALSE;
+            break;
+
+        case RSVG_CSS_COLOR_SPEC_ARGB:
+            state->current_color = spec.argb;
+            state->has_current_color = TRUE;
+            break;
+
+        case RSVG_CSS_COLOR_PARSE_ERROR:
+            /* FIXME: no error handling */
+            state->has_current_color = FALSE;
+            break;
+
+        default:
+            g_assert_not_reached ();
+        }
+    } else if (g_str_equal (name, "opacity")) {
+        RsvgOpacitySpec spec;
+
+        spec = rsvg_css_parse_opacity (value);
+        if (spec.kind == RSVG_OPACITY_SPECIFIED) {
+            state->opacity = spec.opacity;
+        } else {
+            state->opacity = 0;
+            /* FIXME: handle INHERIT and PARSE_ERROR */
+        }
+    } else if (g_str_equal (name, "flood-color")) {
+        RsvgCssColorSpec spec;
+
+        spec = rsvg_css_parse_color (value, ALLOW_INHERIT_YES, ALLOW_CURRENT_COLOR_YES);
+        switch (spec.kind) {
+        case RSVG_CSS_COLOR_SPEC_INHERIT:
+            /* FIXME: we should inherit; see how stop-color is handled in rsvg-styles.c */
+            state->has_current_color = FALSE;
+            break;
+
+        case RSVG_CSS_COLOR_SPEC_CURRENT_COLOR:
+            /* FIXME: in the caller, fix up the current color */
+            state->has_flood_color = FALSE;
+            break;
+
+        case RSVG_CSS_COLOR_SPEC_ARGB:
+            state->flood_color = spec.argb;
+            state->has_flood_color = TRUE;
+            break;
+
+        case RSVG_CSS_COLOR_PARSE_ERROR:
+            /* FIXME: no error handling */
+            state->has_current_color = FALSE;
+            break;
+
+        default:
+            g_assert_not_reached ();
+        }
+    } else if (g_str_equal (name, "flood-opacity")) {
+        RsvgOpacitySpec spec;
+
+        spec = rsvg_css_parse_opacity (value);
+        if (spec.kind == RSVG_OPACITY_SPECIFIED) {
+            state->flood_opacity = spec.opacity;
+        } else {
+            state->flood_opacity = 0;
+            /* FIXME: handle INHERIT and PARSE_ERROR */
+        }
+
         state->has_flood_opacity = TRUE;
     } else if (g_str_equal (name, "filter")) {
         g_free (state->filter);
-        state->filter = rsvg_get_url_string (value);
+        state->filter = rsvg_get_url_string (value, NULL);
     } else if (g_str_equal (name, "mask")) {
         g_free (state->mask);
-        state->mask = rsvg_get_url_string (value);
+        state->mask = rsvg_get_url_string (value, NULL);
     } else if (g_str_equal (name, "baseline-shift")) {
         /* These values come from Inkscape's SP_CSS_BASELINE_SHIFT_(SUB/SUPER/BASELINE);
          * see sp_style_merge_baseline_shift_from_parent()
@@ -576,7 +635,7 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
         }
     } else if (g_str_equal (name, "clip-path")) {
         g_free (state->clip_path);
-        state->clip_path = rsvg_get_url_string (value);
+        state->clip_path = rsvg_get_url_string (value, NULL);
     } else if (g_str_equal (name, "overflow")) {
         if (!g_str_equal (value, "inherit")) {
             state->overflow = rsvg_css_parse_overflow (value, &state->has_overflow);
@@ -667,7 +726,16 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
             rsvg_paint_server_parse (&state->has_fill_server, value);
         rsvg_paint_server_unref (fill);
     } else if (g_str_equal (name, "fill-opacity")) {
-        state->fill_opacity = rsvg_css_parse_opacity (value);
+        RsvgOpacitySpec spec;
+
+        spec = rsvg_css_parse_opacity (value);
+        if (spec.kind == RSVG_OPACITY_SPECIFIED) {
+            state->fill_opacity = spec.opacity;
+        } else {
+            state->fill_opacity = 0;
+            /* FIXME: handle INHERIT and PARSE_ERROR */
+        }
+
         state->has_fill_opacity = TRUE;
     } else if (g_str_equal (name, "fill-rule")) {
         state->has_fill_rule = TRUE;
@@ -706,7 +774,16 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
         else
             g_warning (_("unknown line cap style %s\n"), value);
     } else if (g_str_equal (name, "stroke-opacity")) {
-        state->stroke_opacity = rsvg_css_parse_opacity (value);
+        RsvgOpacitySpec spec;
+
+        spec = rsvg_css_parse_opacity (value);
+        if (spec.kind == RSVG_OPACITY_SPECIFIED) {
+            state->stroke_opacity = spec.opacity;
+        } else {
+            state->stroke_opacity = 0;
+            /* FIXME: handle INHERIT and PARSE_ERROR */
+        }
+
         state->has_stroke_opacity = TRUE;
     } else if (g_str_equal (name, "stroke-linejoin")) {
         state->has_join = TRUE;
@@ -809,54 +886,38 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
 	state->letter_spacing = rsvg_length_parse (value, LENGTH_DIR_HORIZONTAL);
     } else if (g_str_equal (name, "stop-color")) {
         state->has_stop_color = TRUE;
-        if (g_str_equal (value, "inherit")) {
-            state->stop_color_mode = STOP_COLOR_INHERIT;
-            state->has_stop_color = FALSE;
-        } else if (g_str_equal (value, "currentColor")) {
-            state->stop_color_mode = STOP_COLOR_CURRENT_COLOR;
-        } else {
-            state->stop_color = rsvg_css_parse_color (value, &state->has_stop_color);
-            if (state->has_stop_color) {
-                state->stop_color_mode = STOP_COLOR_SPECIFIED;
-            }
-        }
+        state->stop_color = rsvg_css_parse_color (value, ALLOW_INHERIT_YES, ALLOW_CURRENT_COLOR_YES);
     } else if (g_str_equal (name, "stop-opacity")) {
+        state->stop_opacity = rsvg_css_parse_opacity (value);
         state->has_stop_opacity = TRUE;
-        if (g_str_equal (value, "inherit")) {
-            state->has_stop_opacity = FALSE;
-            state->stop_opacity_mode = STOP_OPACITY_INHERIT;
-        } else {
-            state->stop_opacity = rsvg_css_parse_opacity (value);
-            state->stop_opacity_mode = STOP_OPACITY_SPECIFIED;
-        }
     } else if (g_str_equal (name, "marker-start")) {
         g_free (state->startMarker);
-        state->startMarker = rsvg_get_url_string (value);
+        state->startMarker = rsvg_get_url_string (value, NULL);
         state->has_startMarker = TRUE;
     } else if (g_str_equal (name, "marker-mid")) {
         g_free (state->middleMarker);
-        state->middleMarker = rsvg_get_url_string (value);
+        state->middleMarker = rsvg_get_url_string (value, NULL);
         state->has_middleMarker = TRUE;
     } else if (g_str_equal (name, "marker-end")) {
         g_free (state->endMarker);
-        state->endMarker = rsvg_get_url_string (value);
+        state->endMarker = rsvg_get_url_string (value, NULL);
         state->has_endMarker = TRUE;
     } else if (g_str_equal (name, "marker")) {
         if (!state->has_startMarker) {
             g_free (state->startMarker);
-            state->startMarker = rsvg_get_url_string (value);
+            state->startMarker = rsvg_get_url_string (value, NULL);
             state->has_startMarker = TRUE;
         }
 
         if (!state->has_middleMarker) {
             g_free (state->middleMarker);
-            state->middleMarker = rsvg_get_url_string (value);
+            state->middleMarker = rsvg_get_url_string (value, NULL);
             state->has_middleMarker = TRUE;
         }
 
         if (!state->has_endMarker) {
             g_free (state->endMarker);
-            state->endMarker = rsvg_get_url_string (value);
+            state->endMarker = rsvg_get_url_string (value, NULL);
             state->has_endMarker = TRUE;
         }
     } else if (g_str_equal (name, "stroke-miterlimit")) {
@@ -912,7 +973,7 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
 
                 is_even = (n_dashes % 2 == 0);
                 state->dash.n_dash = (is_even ? n_dashes : n_dashes * 2);
-                state->dash.dash = g_new (double, state->dash.n_dash);
+                state->dash.dash = g_new0 (double, state->dash.n_dash);
 
                 /* TODO: handle negative value == error case */
 
@@ -940,64 +1001,64 @@ rsvg_parse_style_pair (RsvgHandle * ctx,
 }
 
 static void
-rsvg_lookup_parse_style_pair (RsvgHandle * ctx, RsvgState * state,
+rsvg_lookup_parse_style_pair (RsvgState * state,
                               const char *key, RsvgPropertyBag * atts)
 {
     const char *value;
 
     if ((value = rsvg_property_bag_lookup (atts, key)) != NULL)
-        rsvg_parse_style_pair (ctx, state, key, value, FALSE);
+        rsvg_parse_style_pair (state, key, value, FALSE);
 }
 
 /* take a pair of the form (fill="#ff00ff") and parse it as a style */
 void
-rsvg_parse_style_pairs (RsvgHandle * ctx, RsvgState * state, RsvgPropertyBag * atts)
+rsvg_parse_style_pairs (RsvgState * state, RsvgPropertyBag * atts)
 {
-    rsvg_lookup_parse_style_pair (ctx, state, "baseline-shift", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "clip-path", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "clip-rule", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "color", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "direction", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "display", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "enable-background", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "comp-op", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "fill", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "fill-opacity", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "fill-rule", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "filter", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "flood-color", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "flood-opacity", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "font-family", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "font-size", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "font-stretch", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "font-style", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "font-variant", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "font-weight", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "marker-end", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "mask", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "marker-mid", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "marker-start", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "opacity", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "overflow", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "shape-rendering", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stop-color", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stop-opacity", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-dasharray", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-dashoffset", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-linecap", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-linejoin", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-miterlimit", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-opacity", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "stroke-width", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "text-anchor", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "text-decoration", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "unicode-bidi", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "letter-spacing", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "visibility", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "writing-mode", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "xml:lang", atts);
-    rsvg_lookup_parse_style_pair (ctx, state, "xml:space", atts);
+    rsvg_lookup_parse_style_pair (state, "baseline-shift", atts);
+    rsvg_lookup_parse_style_pair (state, "clip-path", atts);
+    rsvg_lookup_parse_style_pair (state, "clip-rule", atts);
+    rsvg_lookup_parse_style_pair (state, "color", atts);
+    rsvg_lookup_parse_style_pair (state, "direction", atts);
+    rsvg_lookup_parse_style_pair (state, "display", atts);
+    rsvg_lookup_parse_style_pair (state, "enable-background", atts);
+    rsvg_lookup_parse_style_pair (state, "comp-op", atts);
+    rsvg_lookup_parse_style_pair (state, "fill", atts);
+    rsvg_lookup_parse_style_pair (state, "fill-opacity", atts);
+    rsvg_lookup_parse_style_pair (state, "fill-rule", atts);
+    rsvg_lookup_parse_style_pair (state, "filter", atts);
+    rsvg_lookup_parse_style_pair (state, "flood-color", atts);
+    rsvg_lookup_parse_style_pair (state, "flood-opacity", atts);
+    rsvg_lookup_parse_style_pair (state, "font-family", atts);
+    rsvg_lookup_parse_style_pair (state, "font-size", atts);
+    rsvg_lookup_parse_style_pair (state, "font-stretch", atts);
+    rsvg_lookup_parse_style_pair (state, "font-style", atts);
+    rsvg_lookup_parse_style_pair (state, "font-variant", atts);
+    rsvg_lookup_parse_style_pair (state, "font-weight", atts);
+    rsvg_lookup_parse_style_pair (state, "marker-end", atts);
+    rsvg_lookup_parse_style_pair (state, "mask", atts);
+    rsvg_lookup_parse_style_pair (state, "marker-mid", atts);
+    rsvg_lookup_parse_style_pair (state, "marker-start", atts);
+    rsvg_lookup_parse_style_pair (state, "opacity", atts);
+    rsvg_lookup_parse_style_pair (state, "overflow", atts);
+    rsvg_lookup_parse_style_pair (state, "shape-rendering", atts);
+    rsvg_lookup_parse_style_pair (state, "stop-color", atts);
+    rsvg_lookup_parse_style_pair (state, "stop-opacity", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-dasharray", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-dashoffset", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-linecap", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-linejoin", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-miterlimit", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-opacity", atts);
+    rsvg_lookup_parse_style_pair (state, "stroke-width", atts);
+    rsvg_lookup_parse_style_pair (state, "text-anchor", atts);
+    rsvg_lookup_parse_style_pair (state, "text-decoration", atts);
+    rsvg_lookup_parse_style_pair (state, "unicode-bidi", atts);
+    rsvg_lookup_parse_style_pair (state, "letter-spacing", atts);
+    rsvg_lookup_parse_style_pair (state, "visibility", atts);
+    rsvg_lookup_parse_style_pair (state, "writing-mode", atts);
+    rsvg_lookup_parse_style_pair (state, "xml:lang", atts);
+    rsvg_lookup_parse_style_pair (state, "xml:space", atts);
 
     {
         /* TODO: this conditional behavior isn't quite correct, and i'm not sure it should reside here */
@@ -1076,7 +1137,7 @@ rsvg_parse_style (RsvgHandle * ctx, RsvgState * state, const char *str)
             g_strfreev(split_list);
 
             if (parse_style_value (second_value, &style_value, &important))
-                rsvg_parse_style_pair (ctx, state,
+                rsvg_parse_style_pair (state,
                                        g_strstrip (first_value),
                                        style_value,
                                        important);
@@ -1301,135 +1362,6 @@ ccss_import_style (CRDocHandler * a_this,
     g_free (mime_type);
 }
 
-/* Parse an SVG transform string into an affine matrix. Reference: SVG
-   working draft dated 1999-07-06, section 8.5. Return TRUE on
-   success. */
-gboolean
-rsvg_parse_transform (cairo_matrix_t *dst, const char *src)
-{
-    int idx;
-    char keyword[32];
-    double args[6];
-    int n_args;
-    guint key_len;
-    cairo_matrix_t affine;
-
-    cairo_matrix_init_identity (dst);
-
-    idx = 0;
-    while (src[idx]) {
-        /* skip initial whitespace */
-        while (g_ascii_isspace (src[idx]))
-            idx++;
-
-        if (src[idx] == '\0')
-            break;
-
-        /* parse keyword */
-        for (key_len = 0; key_len < sizeof (keyword); key_len++) {
-            char c;
-
-            c = src[idx];
-            if (g_ascii_isalpha (c) || c == '-')
-                keyword[key_len] = src[idx++];
-            else
-                break;
-        }
-        if (key_len >= sizeof (keyword))
-            return FALSE;
-        keyword[key_len] = '\0';
-
-        /* skip whitespace */
-        while (g_ascii_isspace (src[idx]))
-            idx++;
-
-        if (src[idx] != '(')
-            return FALSE;
-        idx++;
-
-        for (n_args = 0;; n_args++) {
-            char c;
-            char *end_ptr;
-
-            /* skip whitespace */
-            while (g_ascii_isspace (src[idx]))
-                idx++;
-            c = src[idx];
-            if (g_ascii_isdigit (c) || c == '+' || c == '-' || c == '.') {
-                if (n_args == sizeof (args) / sizeof (args[0]))
-                    return FALSE;       /* too many args */
-                args[n_args] = g_ascii_strtod (src + idx, &end_ptr);
-                idx = end_ptr - src;
-
-                while (g_ascii_isspace (src[idx]))
-                    idx++;
-
-                /* skip optional comma */
-                if (src[idx] == ',')
-                    idx++;
-            } else if (c == ')')
-                break;
-            else
-                return FALSE;
-        }
-        idx++;
-
-        /* ok, have parsed keyword and args, now modify the transform */
-        if (!strcmp (keyword, "matrix")) {
-            if (n_args != 6)
-                return FALSE;
-
-            cairo_matrix_init (&affine, args[0], args[1], args[2], args[3], args[4], args[5]);
-            cairo_matrix_multiply (dst, &affine, dst);
-        } else if (!strcmp (keyword, "translate")) {
-            if (n_args == 1)
-                args[1] = 0;
-            else if (n_args != 2)
-                return FALSE;
-            cairo_matrix_init_translate (&affine, args[0], args[1]);
-            cairo_matrix_multiply (dst, &affine, dst);
-        } else if (!strcmp (keyword, "scale")) {
-            if (n_args == 1)
-                args[1] = args[0];
-            else if (n_args != 2)
-                return FALSE;
-            cairo_matrix_init_scale (&affine, args[0], args[1]);
-            cairo_matrix_multiply (dst, &affine, dst);
-        } else if (!strcmp (keyword, "rotate")) {
-            if (n_args == 1) {
-
-                cairo_matrix_init_rotate (&affine, args[0] * M_PI / 180.);
-                cairo_matrix_multiply (dst, &affine, dst);
-            } else if (n_args == 3) {
-                cairo_matrix_init_translate (&affine, args[1], args[2]);
-                cairo_matrix_multiply (dst, &affine, dst);
-
-                cairo_matrix_init_rotate (&affine, args[0] * M_PI / 180.);
-                cairo_matrix_multiply (dst, &affine, dst);
-
-                cairo_matrix_init_translate (&affine, -args[1], -args[2]);
-                cairo_matrix_multiply (dst, &affine, dst);
-            } else
-                return FALSE;
-        } else if (!strcmp (keyword, "skewX")) {
-            if (n_args != 1)
-                return FALSE;
-            _rsvg_cairo_matrix_init_shear (&affine, args[0]);
-            cairo_matrix_multiply (dst, &affine, dst);
-        } else if (!strcmp (keyword, "skewY")) {
-            if (n_args != 1)
-                return FALSE;
-            _rsvg_cairo_matrix_init_shear (&affine, args[0]);
-            /* transpose the affine, given that we know [1] is zero */
-            affine.yx = affine.xy;
-            affine.xy = 0.;
-            cairo_matrix_multiply (dst, &affine, dst);
-        } else
-            return FALSE;       /* unknown keyword */
-    }
-    return TRUE;
-}
-
 /**
  * rsvg_parse_transform_attr:
  * @ctx: Rsvg context.
@@ -1438,7 +1370,7 @@ rsvg_parse_transform (cairo_matrix_t *dst, const char *src)
  *
  * Parses the transform attribute in @str and applies it to @state.
  **/
-static void
+G_GNUC_WARN_UNUSED_RESULT static gboolean
 rsvg_parse_transform_attr (RsvgHandle * ctx, RsvgState * state, const char *str)
 {
     cairo_matrix_t affine;
@@ -1446,6 +1378,9 @@ rsvg_parse_transform_attr (RsvgHandle * ctx, RsvgState * state, const char *str)
     if (rsvg_parse_transform (&affine, str)) {
         cairo_matrix_multiply (&state->personal_affine, &affine, &state->personal_affine);
         cairo_matrix_multiply (&state->affine, &affine, &state->affine);
+        return TRUE;
+    } else {
+        return FALSE;
     }
 }
 
@@ -1458,7 +1393,7 @@ static void
 apply_style (const gchar *key, StyleValueData *value, gpointer user_data)
 {
     StylesData *data = (StylesData *) user_data;
-    rsvg_parse_style_pair (data->ctx, data->state, key, value->value, value->important);
+    rsvg_parse_style_pair (data->state, key, value->value, value->important);
 }
 
 static gboolean
@@ -1469,7 +1404,7 @@ rsvg_lookup_apply_css_style (RsvgHandle * ctx, const char *target, RsvgState * s
     styles = g_hash_table_lookup (ctx->priv->css_props, target);
 
     if (styles != NULL) {
-        StylesData *data = g_new (StylesData, 1);
+        StylesData *data = g_new0 (StylesData, 1);
         data->ctx = ctx;
         data->state = state;
         g_hash_table_foreach (styles, (GHFunc) apply_style, data);
@@ -1482,7 +1417,7 @@ rsvg_lookup_apply_css_style (RsvgHandle * ctx, const char *target, RsvgState * s
 /**
  * rsvg_parse_style_attrs:
  * @ctx: Rsvg context.
- * @state: Rsvg state
+ * @node: Rsvg node whose state should be modified
  * @tag: (nullable): The SVG tag we're processing (eg: circle, ellipse), optionally %NULL
  * @klazz: (nullable): The space delimited class list, optionally %NULL
  * @atts: Attributes in SAX style.
@@ -1491,17 +1426,20 @@ rsvg_lookup_apply_css_style (RsvgHandle * ctx, const char *target, RsvgState * s
  * stack.
  **/
 void
-rsvg_parse_style_attrs (RsvgHandle * ctx,
-                        RsvgState * state,
+rsvg_parse_style_attrs (RsvgHandle *ctx,
+                        RsvgNode *node,
                         const char *tag, const char *klazz, const char *id, RsvgPropertyBag * atts)
 {
     int i = 0, j = 0;
     char *target = NULL;
     gboolean found = FALSE;
     GString *klazz_list = NULL;
+    RsvgState *state;
+
+    state = rsvg_node_get_state (node);
 
     if (rsvg_property_bag_size (atts) > 0)
-        rsvg_parse_style_pairs (ctx, state, atts);
+        rsvg_parse_style_pairs (state, atts);
 
     /* Try to properly support all of the following, including inheritance:
      * *
@@ -1582,8 +1520,14 @@ rsvg_parse_style_attrs (RsvgHandle * ctx,
 
         if ((value = rsvg_property_bag_lookup (atts, "style")) != NULL)
             rsvg_parse_style (ctx, state, value);
-        if ((value = rsvg_property_bag_lookup (atts, "transform")) != NULL)
-            rsvg_parse_transform_attr (ctx, state, value);
+
+        if ((value = rsvg_property_bag_lookup (atts, "transform")) != NULL) {
+            if (!rsvg_parse_transform_attr (ctx, state, value)) {
+                rsvg_node_set_attribute_parse_error (node,
+                                                     "transform",
+                                                     "Invalid transformation");
+            }
+        }
     }
 }
 
@@ -1759,10 +1703,62 @@ rsvg_state_reinherit_top (RsvgDrawingCtx * ctx, RsvgState * state, int dominate)
 }
 
 void
-rsvg_state_reconstruct (RsvgState * state, RsvgNode * current)
+rsvg_state_reconstruct (RsvgState *state, RsvgNode *current)
 {
+    RsvgNode *currents_parent;
+
     if (current == NULL)
         return;
-    rsvg_state_reconstruct (state, rsvg_node_get_parent (current));
+
+    currents_parent = rsvg_node_get_parent (current);
+
+    rsvg_state_reconstruct (state, currents_parent);
+
+    currents_parent = rsvg_node_unref (currents_parent);
+
     rsvg_state_inherit (state, rsvg_node_get_state (current));
 }
+
+gboolean
+rsvg_state_is_overflow (RsvgState *state)
+{
+    return state->overflow;
+}
+
+gboolean
+rsvg_state_has_overflow (RsvgState *state)
+{
+    return state->has_overflow;
+}
+
+gboolean
+rsvg_state_get_cond_true (RsvgState *state)
+{
+    return state->cond_true;
+}
+
+RsvgCssColorSpec *
+rsvg_state_get_stop_color (RsvgState *state)
+{
+    if (state->has_stop_color) {
+        return &state->stop_color;
+    } else {
+        return NULL;
+    }
+}
+
+RsvgOpacitySpec *
+rsvg_state_get_stop_opacity (RsvgState *state)
+{
+    if (state->has_stop_opacity) {
+        return &state->stop_opacity;
+    } else {
+        return NULL;
+    }
+}
+
+guint32
+rsvg_state_get_current_color (RsvgState *state)
+{
+    return state->current_color;
+}
