@@ -1,19 +1,24 @@
-extern crate libc;
-extern crate cairo;
-extern crate cairo_sys;
-extern crate glib;
+use ::cairo;
+use ::glib_sys;
+use ::glib::translate::*;
+use ::libc;
 
-use self::glib::translate::*;
+use std::cell::RefCell;
 
-use length::*;
-
-use drawing_ctx;
-use drawing_ctx::RsvgDrawingCtx;
-use drawing_ctx::RsvgNode;
+use cairo::MatrixTrait;
 
 use bbox::*;
+use drawing_ctx;
+use drawing_ctx::RsvgDrawingCtx;
+use handle::RsvgHandle;
+use length::*;
+use node::*;
+use paint_server::*;
+use property_bag;
+use property_bag::*;
+use stop::*;
+use util::*;
 
-use self::cairo::MatrixTrait;
 
 #[derive(Copy, Clone)]
 pub struct ColorStop {
@@ -25,10 +30,11 @@ pub struct ColorStop {
  * ones can be inherited from the gradient referenced by its "fallback" IRI.  We
  * represent these possibly-missing attributes as Option<foo>.
  */
+#[derive(Clone)]
 pub struct GradientCommon {
-    pub obj_bbox: Option<bool>,
+    pub units:    Option<PaintServerUnits>,
     pub affine:   Option<cairo::Matrix>,
-    pub spread:   Option<cairo::enums::Extend>,
+    pub spread:   Option<PaintServerSpread>,
     pub fallback: Option<String>,
     pub stops:    Option<Vec<ColorStop>>
 }
@@ -51,26 +57,48 @@ pub enum GradientVariant {
     }
 }
 
+#[derive(Clone)]
 pub struct Gradient {
     pub common: GradientCommon,
     pub variant: GradientVariant
 }
 
-impl GradientCommon {
-    fn new (obj_bbox: Option<bool>,
-            affine:   Option<cairo::Matrix>,
-            spread:   Option<cairo::enums::Extend>,
-            fallback: Option<String>,
-            stops:    Option<Vec<ColorStop>>) -> GradientCommon {
+impl Default for GradientCommon {
+    fn default () -> GradientCommon {
         GradientCommon {
-            obj_bbox: obj_bbox,
-            affine:   affine,
-            spread:   spread,
-            fallback: fallback,
-            stops:    stops
+            units:    None,
+            affine:   None,
+            spread:   None,
+            fallback: None,
+            stops:    None,
         }
     }
+}
 
+// All of the Gradient's fields are Option<foo> values, because
+// those fields can be omitted in the SVG file.  We need to resolve
+// them to default values, or to fallback values that come from
+// another Gradient.
+//
+// For the fallback case, this would need something like
+//
+//    if self.foo.is_none () { self.foo = fallback.foo; }
+//
+// And for the default case, it would be like
+//    if self.foo.is_none () { self.foo = Some (default_value); }
+//
+// Both can be replaced by
+//
+//    self.foo = self.foo.take ().or (bar);
+//
+// So we define a macro for that.
+macro_rules! fallback_to (
+    ($dest:expr, $default:expr) => (
+        $dest = $dest.take ().or ($default)
+    );
+);
+
+impl GradientCommon {
     fn clone_stops (&self) -> Option<Vec<ColorStop>> {
         if let Some (ref stops) = self.stops {
             Some (stops.clone ())
@@ -80,7 +108,7 @@ impl GradientCommon {
     }
 
     fn is_resolved (&self) -> bool {
-        self.obj_bbox.is_some() &&
+        self.units.is_some() &&
             self.affine.is_some () &&
             self.spread.is_some () &&
             self.stops.is_some ()
@@ -89,24 +117,21 @@ impl GradientCommon {
     fn resolve_from_defaults (&mut self) {
         /* These are per the spec */
 
-        if self.obj_bbox.is_none () { self.obj_bbox = Some (true); }
-        if self.affine.is_none ()   { self.affine   = Some (cairo::Matrix::identity ()); }
-        if self.spread.is_none ()   { self.spread   = Some (cairo::enums::Extend::Pad); }
+        fallback_to! (self.units,  Some (PaintServerUnits::default ()));
+        fallback_to! (self.affine, Some (cairo::Matrix::identity ()));
+        fallback_to! (self.spread, Some (PaintServerSpread::default ()));
+        fallback_to! (self.stops,  Some (Vec::<ColorStop>::new ())); // empty array of color stops
 
         self.fallback = None;
-
-        if self.stops.is_none ()    { self.stops    = Some (Vec::<ColorStop>::new ()); } // empty array of color stops
     }
 
     fn resolve_from_fallback (&mut self, fallback: &GradientCommon) {
-        if self.obj_bbox.is_none () { self.obj_bbox = fallback.obj_bbox; }
-        if self.affine.is_none ()   { self.affine   = fallback.affine; }
-        if self.spread.is_none ()   { self.spread   = fallback.spread; }
-        if self.stops.is_none ()    { self.stops    = fallback.clone_stops (); }
+        fallback_to! (self.units,  fallback.units);
+        fallback_to! (self.affine, fallback.affine);
+        fallback_to! (self.spread, fallback.spread);
+        fallback_to! (self.stops,  fallback.clone_stops ());
 
-        if self.fallback.is_none () {
-            self.fallback = clone_fallback_name (&fallback.fallback);
-        }
+        self.fallback = clone_fallback_name (&fallback.fallback);
     }
 
     fn add_color_stop (&mut self, mut offset: f64, rgba: u32) {
@@ -133,26 +158,6 @@ impl GradientCommon {
     }
 }
 
-fn clone_fallback_name (fallback: &Option<String>) -> Option<String> {
-    if let Some (ref fallback_name) = *fallback {
-        Some (fallback_name.clone ())
-    } else {
-        None
-    }
-}
-
-impl Clone for GradientCommon {
-    fn clone (&self) -> Self {
-        GradientCommon {
-            obj_bbox: self.obj_bbox,
-            affine:   self.affine,
-            spread:   self.spread,
-            fallback: clone_fallback_name (&self.fallback),
-            stops:    self.clone_stops ()
-        }
-    }
-}
-
 impl GradientVariant {
     fn is_resolved (&self) -> bool {
         match *self {
@@ -177,21 +182,23 @@ impl GradientVariant {
         /* These are per the spec */
 
         match *self {
+            // https://www.w3.org/TR/SVG/pservers.html#LinearGradients
             GradientVariant::Linear { ref mut x1, ref mut y1, ref mut x2, ref mut y2 } => {
-                if x1.is_none () { *x1 = Some (RsvgLength::parse ("0%", LengthDir::Horizontal)); }
-                if y1.is_none () { *y1 = Some (RsvgLength::parse ("0%", LengthDir::Vertical)); }
-                if x2.is_none () { *x2 = Some (RsvgLength::parse ("100%", LengthDir::Horizontal)); }
-                if y2.is_none () { *y2 = Some (RsvgLength::parse ("0%", LengthDir::Vertical)); }
+                fallback_to! (*x1, Some (RsvgLength::parse ("0%", LengthDir::Horizontal).unwrap ()));
+                fallback_to! (*y1, Some (RsvgLength::parse ("0%", LengthDir::Vertical).unwrap ()));
+                fallback_to! (*x2, Some (RsvgLength::parse ("100%", LengthDir::Horizontal).unwrap ()));
+                fallback_to! (*y2, Some (RsvgLength::parse ("0%", LengthDir::Vertical).unwrap ()));
             },
 
+            // https://www.w3.org/TR/SVG/pservers.html#RadialGradients
             GradientVariant::Radial { ref mut cx, ref mut cy, ref mut r, ref mut fx, ref mut fy } => {
-                if cx.is_none () { *cx = Some (RsvgLength::parse ("50%", LengthDir::Horizontal)); }
-                if cy.is_none () { *cy = Some (RsvgLength::parse ("50%", LengthDir::Vertical)); }
-                if r.is_none ()  { *r  = Some (RsvgLength::parse ("50%", LengthDir::Both)); }
+                fallback_to! (*cx, Some (RsvgLength::parse ("50%", LengthDir::Horizontal).unwrap ()));
+                fallback_to! (*cy, Some (RsvgLength::parse ("50%", LengthDir::Vertical).unwrap ()));
+                fallback_to! (*r,  Some (RsvgLength::parse ("50%", LengthDir::Both).unwrap ()));
 
                 /* fx and fy fall back to the presentational value of cx and cy */
-                if fx.is_none () { *fx = *cx }
-                if fy.is_none () { *fy = *cy }
+                fallback_to! (*fx, *cx);
+                fallback_to! (*fy, *cy);
             }
         }
     }
@@ -200,20 +207,20 @@ impl GradientVariant {
         match *self {
             GradientVariant::Linear { ref mut x1, ref mut y1, ref mut x2, ref mut y2 } => {
                 if let &GradientVariant::Linear { x1: x1f, y1: y1f, x2: x2f, y2: y2f } = fallback {
-                    if x1.is_none () { *x1 = x1f; }
-                    if y1.is_none () { *y1 = y1f; }
-                    if x2.is_none () { *x2 = x2f; }
-                    if y2.is_none () { *y2 = y2f; }
+                    fallback_to! (*x1, x1f);
+                    fallback_to! (*y1, y1f);
+                    fallback_to! (*x2, x2f);
+                    fallback_to! (*y2, y2f);
                 }
             },
 
             GradientVariant::Radial { ref mut cx, ref mut cy, ref mut r, ref mut fx, ref mut fy } => {
                 if let &GradientVariant::Radial { cx: cxf, cy: cyf, r: rf, fx: fxf, fy: fyf } = fallback {
-                    if cx.is_none () { *cx = cxf; }
-                    if cy.is_none () { *cy = cyf; }
-                    if r.is_none ()  { *r  = rf;  }
-                    if fx.is_none () { *fx = fxf; }
-                    if fy.is_none () { *fy = fyf; }
+                    fallback_to! (*cx, cxf);
+                    fallback_to! (*cy, cyf);
+                    fallback_to! (*r,  rf);
+                    fallback_to! (*fx, fxf);
+                    fallback_to! (*fy, fyf);
                 }
             }
         }
@@ -221,13 +228,6 @@ impl GradientVariant {
 }
 
 impl Gradient {
-    fn new (common: GradientCommon, variant: GradientVariant) -> Gradient {
-        Gradient {
-            common: common,
-            variant: variant
-        }
-    }
-
     fn is_resolved (&self) -> bool {
         self.common.is_resolved () && self.variant.is_resolved ()
     }
@@ -242,6 +242,24 @@ impl Gradient {
         self.variant.resolve_from_fallback (&fallback.variant);
     }
 
+    fn add_color_stops_from_node (&mut self, node: &RsvgNode) {
+        assert! (node.get_type () == NodeType::LinearGradient || node.get_type () == NodeType::RadialGradient);
+
+        for child in &*node.children.borrow () {
+            if child.get_type () != NodeType::Stop {
+                continue; // just ignore this child; we are only interested in gradient stops
+            }
+
+            if child.get_result ().is_err () {
+                break; // don't add any more stops
+            }
+
+            child.with_impl (|stop: &NodeStop| {
+                self.add_color_stop (stop.get_offset (), stop.get_rgba ());
+            });
+        }
+    }
+
     fn add_color_stop (&mut self, offset: f64, rgba: u32) {
         self.common.add_color_stop (offset, rgba);
     }
@@ -262,31 +280,25 @@ impl Gradient {
     }
 }
 
-impl Clone for Gradient {
-    fn clone (&self) -> Self {
-        Gradient {
-            common: self.common.clone (),
-            variant: self.variant
-        }
-    }
-}
-
 trait FallbackSource {
-    fn get_fallback (&mut self, name: &str) -> Option<Box<Gradient>>;
+    fn get_fallback (&mut self, name: &str) -> Option<RsvgNode>;
 }
 
 fn resolve_gradient (gradient: &Gradient, fallback_source: &mut FallbackSource) -> Gradient {
     let mut result = gradient.clone ();
 
     while !result.is_resolved () {
-        let mut opt_fallback: Option<Box<Gradient>> = None;
+        let mut opt_fallback: Option<RsvgNode> = None;
 
         if let Some (ref fallback_name) = result.common.fallback {
             opt_fallback = fallback_source.get_fallback (&**fallback_name);
         }
 
-        if let Some (fallback_gradient) = opt_fallback {
-            result.resolve_from_fallback (&*fallback_gradient);
+        if let Some (fallback_node) = opt_fallback {
+            fallback_node.with_impl (|i: &NodeGradient| {
+                let fallback_gradient = i.get_gradient_with_color_stops_from_node (&fallback_node);
+                result.resolve_from_fallback (&fallback_gradient)
+            });
         } else {
             result.resolve_from_defaults ();
             break;
@@ -318,29 +330,22 @@ impl Drop for NodeFallbackSource {
     }
 }
 
-extern "C" {
-    fn rsvg_gradient_node_to_rust_gradient (node: *const RsvgNode) -> *mut Gradient;
-}
-
 impl FallbackSource for NodeFallbackSource {
-    fn get_fallback (&mut self, name: &str) -> Option<Box<Gradient>> {
+    fn get_fallback (&mut self, name: &str) -> Option<RsvgNode> {
         let fallback_node = drawing_ctx::acquire_node (self.draw_ctx, name);
 
         if fallback_node.is_null () {
             return None;
         }
 
-        self.acquired_nodes.push (fallback_node);
-
-        let raw_fallback_gradient = unsafe { rsvg_gradient_node_to_rust_gradient (fallback_node) };
-
-        if raw_fallback_gradient.is_null () {
+        let node: &RsvgNode = unsafe { & *fallback_node };
+        if !(node.get_type () == NodeType::LinearGradient || node.get_type () == NodeType::RadialGradient) {
             return None;
         }
 
-        let fallback_gradient = unsafe { Box::from_raw (raw_fallback_gradient) };
+        self.acquired_nodes.push (fallback_node);
 
-        return Some (fallback_gradient);
+        return Some (node.clone ());
     }
 }
 
@@ -354,9 +359,9 @@ fn set_common_on_pattern<P: cairo::Pattern + cairo::Gradient> (gradient: &Gradie
 
     let mut affine = gradient.common.affine.unwrap ();
 
-    let obj_bbox = gradient.common.obj_bbox.unwrap ();
+    let units = gradient.common.units.unwrap ();
 
-    if obj_bbox {
+    if units == PaintServerUnits::ObjectBoundingBox {
         let bbox_matrix = cairo::Matrix::new (bbox.rect.width, 0.0,
                                               0.0, bbox.rect.height,
                                               bbox.rect.x, bbox.rect.y);
@@ -365,7 +370,7 @@ fn set_common_on_pattern<P: cairo::Pattern + cairo::Gradient> (gradient: &Gradie
 
     affine.invert ();
     pattern.set_matrix (affine);
-    pattern.set_extend (gradient.common.spread.unwrap ());
+    pattern.set_extend (gradient.common.spread.unwrap ().0);
 
     gradient.add_color_stops_to_pattern (pattern, opacity);
 
@@ -375,12 +380,11 @@ fn set_common_on_pattern<P: cairo::Pattern + cairo::Gradient> (gradient: &Gradie
 fn set_linear_gradient_on_pattern (gradient: &Gradient,
                                    draw_ctx: *mut RsvgDrawingCtx,
                                    bbox:     &RsvgBbox,
-                                   opacity:  u8)
-{
+                                   opacity:  u8) -> bool {
     if let GradientVariant::Linear { x1, y1, x2, y2 } = gradient.variant {
-        let obj_bbox = gradient.common.obj_bbox.unwrap ();
+        let units = gradient.common.units.unwrap ();
 
-        if obj_bbox {
+        if units == PaintServerUnits::ObjectBoundingBox {
             drawing_ctx::push_view_box (draw_ctx, 1.0, 1.0);
         }
 
@@ -389,7 +393,7 @@ fn set_linear_gradient_on_pattern (gradient: &Gradient,
                                                       x2.as_ref ().unwrap ().normalize (draw_ctx),
                                                       y2.as_ref ().unwrap ().normalize (draw_ctx));
 
-        if obj_bbox {
+        if units == PaintServerUnits::ObjectBoundingBox {
             drawing_ctx::pop_view_box (draw_ctx);
         }
 
@@ -397,6 +401,8 @@ fn set_linear_gradient_on_pattern (gradient: &Gradient,
     } else {
         unreachable! ();
     }
+
+    true
 }
 
 /* SVG defines radial gradients as being inside a circle (cx, cy, radius).  The
@@ -456,11 +462,11 @@ fn fix_focus_point (mut fx: f64,
 fn set_radial_gradient_on_pattern (gradient: &Gradient,
                                    draw_ctx: *mut RsvgDrawingCtx,
                                    bbox:     &RsvgBbox,
-                                   opacity:  u8) {
+                                   opacity:  u8) -> bool {
     if let GradientVariant::Radial { cx, cy, r, fx, fy } = gradient.variant {
-        let obj_bbox = gradient.common.obj_bbox.unwrap ();
+        let units = gradient.common.units.unwrap ();
 
-        if obj_bbox {
+        if units == PaintServerUnits::ObjectBoundingBox {
             drawing_ctx::push_view_box (draw_ctx, 1.0, 1.0);
         }
 
@@ -474,7 +480,7 @@ fn set_radial_gradient_on_pattern (gradient: &Gradient,
 
         let mut pattern = cairo::RadialGradient::new (new_fx, new_fy, 0.0, n_cx, n_cy, n_r);
 
-        if obj_bbox {
+        if units == PaintServerUnits::ObjectBoundingBox {
             drawing_ctx::pop_view_box (draw_ctx);
         }
 
@@ -482,127 +488,158 @@ fn set_radial_gradient_on_pattern (gradient: &Gradient,
     } else {
         unreachable! ();
     }
+
+    true
 }
 
 fn set_pattern_on_draw_context (gradient: &Gradient,
                                 draw_ctx: *mut RsvgDrawingCtx,
                                 opacity:  u8,
-                                bbox:     &RsvgBbox) {
+                                bbox:     &RsvgBbox) -> bool {
     assert! (gradient.is_resolved ());
 
     match gradient.variant {
         GradientVariant::Linear { .. } => {
-            set_linear_gradient_on_pattern (gradient, draw_ctx, bbox, opacity);
+            set_linear_gradient_on_pattern (gradient, draw_ctx, bbox, opacity)
         }
 
         GradientVariant::Radial { .. } => {
-            set_radial_gradient_on_pattern (gradient, draw_ctx, bbox, opacity);
+            set_radial_gradient_on_pattern (gradient, draw_ctx, bbox, opacity)
+        }
+    }
+}
+
+struct NodeGradient {
+    gradient: RefCell <Gradient>
+}
+
+impl NodeGradient {
+    fn new_linear () -> NodeGradient {
+        NodeGradient {
+            gradient: RefCell::new (Gradient {
+                common: GradientCommon::default (),
+                variant: GradientVariant::Linear {
+                    x1: None,
+                    y1: None,
+                    x2: None,
+                    y2: None
+                }
+            })
         }
     }
 
+    fn new_radial () -> NodeGradient {
+        NodeGradient {
+            gradient: RefCell::new (Gradient {
+                common: GradientCommon::default (),
+                variant: GradientVariant::Radial {
+                    cx: None,
+                    cy: None,
+                    r:  None,
+                    fx: None,
+                    fy: None
+                }
+            })
+        }
+    }
 
+    fn get_gradient_with_color_stops_from_node (&self, node: &RsvgNode) -> Gradient {
+        let mut gradient = self.gradient.borrow ().clone ();
+        gradient.add_color_stops_from_node (node);
+        gradient
+    }
 }
 
+impl NodeTrait for NodeGradient {
+    fn set_atts (&self, node: &RsvgNode, _: *const RsvgHandle, pbag: *const RsvgPropertyBag) -> NodeResult {
+        let mut g = self.gradient.borrow_mut ();
 
-/* All the arguments are pointers because they are in fact optional in
- * SVG.  We turn the arguments into Option<foo>: NULL into None, and
- * anything else into a Some().
- */
-#[no_mangle]
-pub unsafe extern fn gradient_linear_new (x1: *const RsvgLength,
-                                          y1: *const RsvgLength,
-                                          x2: *const RsvgLength,
-                                          y2: *const RsvgLength,
-                                          obj_bbox: *const bool,
-                                          affine: *const cairo::Matrix,
-                                          spread: *const cairo::enums::Extend,
-                                          fallback_name: *const libc::c_char) -> *mut Gradient {
-    let my_obj_bbox      = { if obj_bbox.is_null ()      { None } else { Some (*obj_bbox) } };
-    let my_affine        = { if affine.is_null ()        { None } else { Some (*affine) } };
-    let my_spread        = { if spread.is_null ()        { None } else { Some (*spread) } };
-    let my_fallback_name = { if fallback_name.is_null () { None } else { Some (String::from_glib_none (fallback_name)) } };
+        // Attributes common to linear and radial gradients
 
-    let my_x1 = { if x1.is_null () { None } else { Some (*x1) } };
-    let my_y1 = { if y1.is_null () { None } else { Some (*y1) } };
-    let my_x2 = { if x2.is_null () { None } else { Some (*x2) } };
-    let my_y2 = { if y2.is_null () { None } else { Some (*y2) } };
+        g.common.units    = property_bag::parse_or_none (pbag, "gradientUnits")?;
+        g.common.affine   = property_bag::transform_or_none (pbag, "gradientTransform")?;
+        g.common.spread   = property_bag::parse_or_none (pbag, "spreadMethod")?;
+        g.common.fallback = property_bag::lookup (pbag, "xlink:href");
 
-    let gradient = Gradient::new (GradientCommon::new (my_obj_bbox, my_affine, my_spread, my_fallback_name, None),
-                                  GradientVariant::Linear { x1: my_x1,
-                                                            y1: my_y1,
-                                                            x2: my_x2,
-                                                            y2: my_y2 });
+        // Attributes specific to each gradient type.  The defaults mandated by the spec
+        // are in GradientVariant::resolve_from_defaults()
 
-    let boxed_gradient = Box::new (gradient);
+        match node.get_type () {
+            NodeType::LinearGradient => {
+                g.variant = GradientVariant::Linear {
+                    x1: property_bag::length_or_none (pbag, "x1", LengthDir::Horizontal)?,
+                    y1: property_bag::length_or_none (pbag, "y1", LengthDir::Vertical)?,
+                    x2: property_bag::length_or_none (pbag, "x2", LengthDir::Horizontal)?,
+                    y2: property_bag::length_or_none (pbag, "y2", LengthDir::Vertical)?
+                };
+            },
 
-    Box::into_raw (boxed_gradient)
+            NodeType::RadialGradient => {
+                g.variant = GradientVariant::Radial {
+                    cx: property_bag::length_or_none (pbag, "cx", LengthDir::Horizontal)?,
+                    cy: property_bag::length_or_none (pbag, "cy", LengthDir::Vertical)?,
+                    r:  property_bag::length_or_none (pbag, "r",  LengthDir::Both)?,
+                    fx: property_bag::length_or_none (pbag, "fx", LengthDir::Horizontal)?,
+                    fy: property_bag::length_or_none (pbag, "fy", LengthDir::Vertical)?
+                };
+            },
+
+            _ => unreachable! ()
+        }
+
+        Ok (())
+    }
+
+    fn draw (&self, _: &RsvgNode, _: *const RsvgDrawingCtx, _: i32) {
+        // nothing; paint servers are handled specially
+    }
+
+    fn get_c_impl (&self) -> *const RsvgCNodeImpl {
+        unreachable! ();
+    }
 }
 
 #[no_mangle]
-pub unsafe extern fn gradient_radial_new (cx: *const RsvgLength,
-                                          cy: *const RsvgLength,
-                                          r:  *const RsvgLength,
-                                          fx: *const RsvgLength,
-                                          fy: *const RsvgLength,
-                                          obj_bbox: *const bool,
-                                          affine: *const cairo::Matrix,
-                                          spread: *const cairo::enums::Extend,
-                                          fallback_name: *const libc::c_char) -> *mut Gradient {
-    let my_obj_bbox      = { if obj_bbox.is_null ()      { None } else { Some (*obj_bbox) } };
-    let my_affine        = { if affine.is_null ()        { None } else { Some (*affine) } };
-    let my_spread        = { if spread.is_null ()        { None } else { Some (*spread) } };
-    let my_fallback_name = { if fallback_name.is_null () { None } else { Some (String::from_glib_none (fallback_name)) } };
-
-    let my_cx = { if cx.is_null () { None } else { Some (*cx) } };
-    let my_cy = { if cy.is_null () { None } else { Some (*cy) } };
-    let my_r  = { if r.is_null  () { None } else { Some (*r)  } };
-    let my_fx = { if fx.is_null () { None } else { Some (*fx) } };
-    let my_fy = { if fy.is_null () { None } else { Some (*fy) } };
-
-    let gradient = Gradient::new (GradientCommon::new (my_obj_bbox, my_affine, my_spread, my_fallback_name, None),
-                                  GradientVariant::Radial { cx: my_cx,
-                                                            cy: my_cy,
-                                                            r:  my_r,
-                                                            fx: my_fx,
-                                                            fy: my_fy });
-
-    let boxed_gradient = Box::new (gradient);
-
-    Box::into_raw (boxed_gradient)
+pub extern fn rsvg_node_linear_gradient_new (_: *const libc::c_char, raw_parent: *const RsvgNode) -> *const RsvgNode {
+    boxed_node_new (NodeType::LinearGradient,
+                    raw_parent,
+                    Box::new (NodeGradient::new_linear ()))
 }
 
 #[no_mangle]
-pub unsafe extern fn gradient_destroy (raw_gradient: *mut Gradient) {
-    assert! (!raw_gradient.is_null ());
-
-    let _ = Box::from_raw (raw_gradient);
+pub extern fn rsvg_node_radial_gradient_new (_: *const libc::c_char, raw_parent: *const RsvgNode) -> *const RsvgNode {
+    boxed_node_new (NodeType::RadialGradient,
+                    raw_parent,
+                    Box::new (NodeGradient::new_radial ()))
 }
 
-#[no_mangle]
-pub extern fn gradient_add_color_stop (raw_gradient: *mut Gradient,
-                                       offset:       f64,
-                                       rgba:         u32) {
-    assert! (!raw_gradient.is_null ());
-
-    let gradient: &mut Gradient = unsafe { &mut (*raw_gradient) };
-
-    gradient.add_color_stop (offset, rgba);
-}
-
-#[no_mangle]
-pub extern fn gradient_resolve_fallbacks_and_set_pattern (raw_gradient: *mut Gradient,
-                                                          draw_ctx:     *mut RsvgDrawingCtx,
-                                                          opacity:      u8,
-                                                          bbox:         RsvgBbox) {
-    assert! (!raw_gradient.is_null ());
-    let gradient: &mut Gradient = unsafe { &mut (*raw_gradient) };
-
+fn resolve_fallbacks_and_set_pattern (gradient: &Gradient,
+                                      draw_ctx: *mut RsvgDrawingCtx,
+                                      opacity:  u8,
+                                      bbox:     RsvgBbox) -> bool {
     let mut fallback_source = NodeFallbackSource::new (draw_ctx);
 
     let resolved = resolve_gradient (gradient, &mut fallback_source);
 
-    set_pattern_on_draw_context (&resolved,
-                                 draw_ctx,
-                                 opacity,
-                                 &bbox);
+    set_pattern_on_draw_context (&resolved, draw_ctx, opacity, &bbox)
+}
+
+#[no_mangle]
+pub extern fn gradient_resolve_fallbacks_and_set_pattern (raw_node:     *const RsvgNode,
+                                                          draw_ctx:     *mut RsvgDrawingCtx,
+                                                          opacity:      u8,
+                                                          bbox:         RsvgBbox) -> glib_sys::gboolean {
+    assert! (!raw_node.is_null ());
+    let node: &RsvgNode = unsafe { & *raw_node };
+
+    assert! (node.get_type () == NodeType::LinearGradient || node.get_type () == NodeType::RadialGradient);
+
+    let mut did_set_gradient = false;
+
+    node.with_impl (|node_gradient: &NodeGradient| {
+        let gradient = node_gradient.get_gradient_with_color_stops_from_node (&node);
+        did_set_gradient = resolve_fallbacks_and_set_pattern (&gradient, draw_ctx, opacity, bbox);
+    });
+
+    did_set_gradient.to_glib ()
 }
