@@ -1,10 +1,230 @@
-extern crate libc;
-extern crate cairo;
-extern crate cairo_sys;
+use ::cairo;
+use ::glib::translate::*;
+use ::libc;
 
-use std::f64;
-use path_builder::*;
+use std::cell::Cell;
+use std::f64::consts::*;
+use std::str::FromStr;
+
+use cairo::MatrixTrait;
+
+use aspect_ratio::*;
+use drawing_ctx;
 use drawing_ctx::RsvgDrawingCtx;
+use error::*;
+use handle::RsvgHandle;
+use length::*;
+use node::*;
+use path_builder::*;
+use parsers;
+use parsers::ParseError;
+use property_bag;
+use property_bag::*;
+use util::*;
+use viewbox::*;
+
+// markerUnits attribute: https://www.w3.org/TR/SVG/painting.html#MarkerElement
+
+#[derive(Debug, Copy, Clone, PartialEq)]
+enum MarkerUnits {
+    UserSpaceOnUse,
+    StrokeWidth
+}
+
+impl Default for MarkerUnits {
+    fn default () -> MarkerUnits {
+        MarkerUnits::StrokeWidth
+    }
+}
+
+impl FromStr for MarkerUnits {
+    type Err = AttributeError;
+
+    fn from_str (s: &str) -> Result <MarkerUnits, AttributeError> {
+        match s {
+            "userSpaceOnUse" => Ok (MarkerUnits::UserSpaceOnUse),
+            "strokeWidth"    => Ok (MarkerUnits::StrokeWidth),
+            _                => Err (AttributeError::Parse (ParseError::new ("expected \"userSpaceOnUse\" or \"strokeWidth\"")))
+        }
+    }
+}
+
+// orient attribute: https://www.w3.org/TR/SVG/painting.html#MarkerElement
+
+#[derive(Debug, Copy, Clone, PartialEq)]
+enum MarkerOrient {
+    Auto,
+    Degrees (f64)
+}
+
+impl Default for MarkerOrient {
+    fn default () -> MarkerOrient {
+        MarkerOrient::Degrees (0.0)
+    }
+}
+
+impl FromStr for MarkerOrient {
+    type Err = AttributeError;
+
+    fn from_str (s: &str) -> Result <MarkerOrient, AttributeError> {
+        match s {
+            "auto" => Ok (MarkerOrient::Auto),
+            _      => parsers::angle_degrees (s)
+                .map (|degrees| MarkerOrient::Degrees (degrees) )
+                .map_err (|e| AttributeError::Parse (e))
+        }
+    }
+}
+
+// NodeMarker
+
+struct NodeMarker {
+    units:  Cell<MarkerUnits>,
+    ref_x:  Cell<RsvgLength>,
+    ref_y:  Cell<RsvgLength>,
+    width:  Cell<RsvgLength>,
+    height: Cell<RsvgLength>,
+    orient: Cell<MarkerOrient>,
+    aspect: Cell<AspectRatio>,
+    vbox:   Cell<Option<ViewBox>>
+}
+
+impl NodeMarker {
+    fn new () -> NodeMarker {
+        NodeMarker {
+            units:  Cell::new (MarkerUnits::default ()),
+            ref_x:  Cell::new (RsvgLength::default ()),
+            ref_y:  Cell::new (RsvgLength::default ()),
+            width:  Cell::new (NodeMarker::get_default_size ()),
+            height: Cell::new (NodeMarker::get_default_size ()),
+            orient: Cell::new (MarkerOrient::default ()),
+            aspect: Cell::new (AspectRatio::default ()),
+            vbox:   Cell::new (None)
+        }
+    }
+
+    fn get_default_size () -> RsvgLength {
+        // per the spec
+        RsvgLength::parse ("3", LengthDir::Both).unwrap ()
+    }
+
+    fn render (&self,
+               node:           &RsvgNode,
+               c_node:         *const RsvgNode,
+               draw_ctx:       *const RsvgDrawingCtx,
+               xpos:           f64,
+               ypos:           f64,
+               computed_angle: f64,
+               line_width:     f64) {
+        let marker_width = self.width.get ().normalize (draw_ctx);
+        let marker_height = self.height.get ().normalize (draw_ctx);
+
+        let mut affine = cairo::Matrix::identity ();
+        affine.translate (xpos, ypos);
+
+        affine = cairo::Matrix::multiply (&affine, &drawing_ctx::get_current_state_affine (draw_ctx));
+
+        let rotation: f64;
+
+        match self.orient.get () {
+            MarkerOrient::Auto =>        { rotation = computed_angle; },
+            MarkerOrient::Degrees (d) => { rotation = d * PI / 180.0; }
+        }
+
+        affine.rotate (rotation);
+
+        if self.units.get () == MarkerUnits::StrokeWidth {
+            affine.scale (line_width, line_width);
+        }
+
+        if let Some (vbox) = self.vbox.get () {
+            let (_, _, w, h) = self.aspect.get ().compute (vbox.0.width, vbox.0.height,
+                                                           0.0, 0.0,
+                                                           marker_width, marker_height);
+
+            affine.scale (w / vbox.0.width, h / vbox.0.height);
+
+            drawing_ctx::push_view_box (draw_ctx, vbox.0.width, vbox.0.height);
+        } else {
+            drawing_ctx::push_view_box (draw_ctx, marker_width, marker_height);
+        }
+
+        affine.translate (-self.ref_x.get ().normalize (draw_ctx),
+                          -self.ref_y.get ().normalize (draw_ctx));
+
+        drawing_ctx::state_push (draw_ctx);
+
+        let state = drawing_ctx::get_current_state (draw_ctx);
+        drawing_ctx::state_reinit (state);
+        drawing_ctx::state_reconstruct (state, c_node);
+
+        drawing_ctx::set_current_state_affine (draw_ctx, affine);
+
+        drawing_ctx::push_discrete_layer (draw_ctx);
+
+        let state = drawing_ctx::get_current_state (draw_ctx);
+
+        if !drawing_ctx::state_is_overflow (state) {
+            if let Some (vbox) = self.vbox.get () {
+                drawing_ctx::add_clipping_rect (draw_ctx,
+                                                vbox.0.x,
+                                                vbox.0.y,
+                                                vbox.0.width,
+                                                vbox.0.height);
+            } else {
+                drawing_ctx::add_clipping_rect (draw_ctx,
+                                                0.0,
+                                                0.0,
+                                                marker_width,
+                                                marker_height);
+            }
+        }
+
+        node.draw_children (draw_ctx, -1); // dominate=-1 so it won't reinherit state / push a layer
+
+        drawing_ctx::state_pop (draw_ctx);
+        drawing_ctx::pop_discrete_layer (draw_ctx);
+        drawing_ctx::pop_view_box (draw_ctx);
+    }
+}
+
+impl NodeTrait for NodeMarker {
+    fn set_atts (&self, _: &RsvgNode, _: *const RsvgHandle, pbag: *const RsvgPropertyBag) -> NodeResult {
+        self.units.set (property_bag::parse_or_default (pbag, "markerUnits")?);
+
+        self.ref_x.set (property_bag::length_or_default (pbag, "refX", LengthDir::Horizontal)?);
+        self.ref_y.set (property_bag::length_or_default (pbag, "refY", LengthDir::Vertical)?);
+
+        self.width.set (property_bag::lookup (pbag, "markerWidth").map_or (NodeMarker::get_default_size (),
+                                                                           |v| RsvgLength::parse (&v, LengthDir::Horizontal).unwrap_or (NodeMarker::get_default_size ())));
+        self.height.set (property_bag::lookup (pbag, "markerHeight").map_or (NodeMarker::get_default_size (),
+                                                                             |v| RsvgLength::parse (&v, LengthDir::Vertical).unwrap_or (NodeMarker::get_default_size ())));
+
+        self.orient.set (property_bag::parse_or_default (pbag, "orient")?);
+        self.aspect.set (property_bag::parse_or_default (pbag, "preserveAspectRatio")?);
+        self.vbox.set   (property_bag::parse_or_none (pbag, "viewBox")?);
+        self.aspect.set (property_bag::parse_or_default (pbag, "preserveAspectRatio")?);
+
+        Ok (())
+    }
+
+    fn draw (&self, _: &RsvgNode, _: *const RsvgDrawingCtx, _: i32) {
+        // nothing; markers are drawn by their referencing shapes
+    }
+
+    fn get_c_impl (&self) -> *const RsvgCNodeImpl {
+        unreachable! ();
+    }
+}
+
+#[no_mangle]
+pub extern fn rsvg_node_marker_new (_: *const libc::c_char, raw_parent: *const RsvgNode) -> *const RsvgNode {
+    boxed_node_new (NodeType::Marker,
+                    raw_parent,
+                    Box::new (NodeMarker::new ()))
+}
+
+// Machinery to figure out marker orientations
 
 #[derive(Debug, PartialEq)]
 pub enum Segment {
@@ -45,12 +265,6 @@ enum SegmentState {
  *    The tangents also work if the segment refers to a lineto (they will both just point in the same direction).
  */
 
-const EPSILON: f64 = 1e-10;
-
-fn double_equals (a: f64, b: f64) -> bool {
-    (a - b).abs () < EPSILON
-}
-
 fn make_degenerate (x: f64, y: f64) -> Segment {
     Segment::Degenerate { x: x, y: y }
 }
@@ -90,12 +304,12 @@ pub fn path_builder_to_segments (builder: &RsvgPathBuilder) -> Vec<Segment> {
     segments = Vec::new ();
     state = SegmentState::Initial;
 
-    for cairo_segment in builder.get_path_segments () {
+    for path_command in builder.get_path_commands () {
         last_x = cur_x;
         last_y = cur_y;
 
-        match *cairo_segment {
-            cairo::PathSegment::MoveTo ((x, y)) => {
+        match *path_command {
+            PathCommand::MoveTo (x, y) => {
                 cur_x = x;
                 cur_y = y;
 
@@ -135,7 +349,7 @@ pub fn path_builder_to_segments (builder: &RsvgPathBuilder) -> Vec<Segment> {
                 }
             },
 
-            cairo::PathSegment::LineTo ((x, y)) => {
+            PathCommand::LineTo (x, y) => {
                 cur_x = x;
                 cur_y = y;
 
@@ -144,7 +358,7 @@ pub fn path_builder_to_segments (builder: &RsvgPathBuilder) -> Vec<Segment> {
                 state = SegmentState::InSubpath;
             },
 
-            cairo::PathSegment::CurveTo ((x2, y2), (x3, y3), (x4, y4)) => {
+            PathCommand::CurveTo ((x2, y2), (x3, y3), (x4, y4)) => {
                 cur_x = x4;
                 cur_y = y4;
 
@@ -153,7 +367,7 @@ pub fn path_builder_to_segments (builder: &RsvgPathBuilder) -> Vec<Segment> {
                 state = SegmentState::InSubpath;
             },
 
-            cairo::PathSegment::ClosePath => {
+            PathCommand::ClosePath => {
                 cur_x = subpath_start_x;
                 cur_y = subpath_start_y;
 
@@ -302,104 +516,150 @@ fn find_outgoing_directionality_forwards (segments: &[Segment], start_index: usi
     (false, 0.0, 0.0)
 }
 
+// Normalizes an angle to [0.0, 2*PI)
+fn normalize_angle (mut angle: f64) -> f64 {
+    if angle < 0.0 {
+        while angle < 0.0 {
+            angle += PI * 2.0;
+        }
+    } else {
+        while angle > PI * 2.0 {
+            angle -= PI * 2.0;
+        }
+    }
+
+    angle
+}
+
 fn angle_from_vector (vx: f64, vy: f64) -> f64 {
-    let mut angle = vy.atan2 (vx);
+    let angle = vy.atan2 (vx);
 
     if angle.is_nan () {
         0.0
     } else {
-        while angle < 0.0 {
-            angle += f64::consts::PI * 2.0;
-        }
-
-        angle
+        normalize_angle (angle)
     }
 }
 
+fn bisect_angles (incoming: f64, outgoing: f64) -> f64 {
+    let half_delta: f64;
+
+    half_delta = (outgoing - incoming) * 0.5;
+
+    if FRAC_PI_2 < half_delta.abs () {
+        normalize_angle (incoming + half_delta - PI)
+    } else {
+        normalize_angle (incoming + half_delta)
+    }
+}
+
+// From SVG's marker-start, marker-mid, marker-end properties
+#[derive(Debug, PartialEq)]
+enum MarkerType {
+    Start,
+    Middle,
+    End
+}
+
+fn emit_marker_by_name (draw_ctx:       *const RsvgDrawingCtx,
+                        marker_name:    *const libc::c_char,
+                        xpos:           f64,
+                        ypos:           f64,
+                        computed_angle: f64,
+                        line_width:     f64) {
+    if marker_name.is_null () {
+        return;
+    }
+
+    let name = unsafe { String::from_glib_none (marker_name) };
+
+    let c_node = drawing_ctx::acquire_node_of_type (draw_ctx, &name, NodeType::Marker);
+
+    if c_node.is_null () {
+        return;
+    }
+
+    let node: &RsvgNode = unsafe { & *c_node };
+
+    node.with_impl (|marker: &NodeMarker| marker.render (node, c_node, draw_ctx, xpos, ypos, computed_angle, line_width));
+
+    drawing_ctx::release_node (draw_ctx, c_node);
+}
+
+fn get_marker_name_from_drawing_ctx (draw_ctx:    *const RsvgDrawingCtx,
+                                     marker_type: MarkerType) -> *const libc::c_char {
+   match marker_type {
+        MarkerType::Start  => unsafe { rsvg_get_start_marker (draw_ctx) },
+        MarkerType::Middle => unsafe { rsvg_get_middle_marker (draw_ctx) },
+        MarkerType::End    => unsafe { rsvg_get_end_marker (draw_ctx) }
+    }
+}
+
+enum MarkerEndpoint {
+    Start,
+    End
+}
+
+fn emit_marker<E> (segment:     &Segment,
+                   endpoint:    MarkerEndpoint,
+                   marker_type: MarkerType,
+                   orient:      f64,
+                   emit_fn:     &mut E) where E: FnMut(MarkerType, f64, f64, f64) {
+    let (x, y) = match *segment {
+        Segment::Degenerate  { x, y } => (x, y),
+
+        Segment::LineOrCurve { x1, y1, x4, y4, .. } => match endpoint {
+            MarkerEndpoint::Start => (x1, y1),
+            MarkerEndpoint::End   => (x4, y4)
+        }
+    };
+
+    emit_fn (marker_type, x, y, orient);
+}
+
 extern "C" {
-    fn rsvg_marker_render (marker_name: *const libc::c_char,
-                           xpos: f64,
-                           ypos: f64,
-                           orient: f64,
-                           linewidth: f64,
-                           ctx: *mut RsvgDrawingCtx);
-    fn rsvg_get_normalized_stroke_width (ctx: *const RsvgDrawingCtx) -> f64;
+    fn rsvg_get_normalized_stroke_width (draw_ctx: *const RsvgDrawingCtx) -> f64;
 
-    fn rsvg_get_start_marker (ctx: *const RsvgDrawingCtx) -> *const libc::c_char;
-    fn rsvg_get_middle_marker (ctx: *const RsvgDrawingCtx) -> *const libc::c_char;
-    fn rsvg_get_end_marker (ctx: *const RsvgDrawingCtx) -> *const libc::c_char;
+    fn rsvg_get_start_marker (draw_ctx: *const RsvgDrawingCtx) -> *const libc::c_char;
+    fn rsvg_get_middle_marker (draw_ctx: *const RsvgDrawingCtx) -> *const libc::c_char;
+    fn rsvg_get_end_marker (draw_ctx: *const RsvgDrawingCtx) -> *const libc::c_char;
 }
 
-enum SubpathState {
-    NoSubpath,
-    InSubpath
+fn drawing_ctx_has_markers (draw_ctx: *const RsvgDrawingCtx) -> bool {
+    (!get_marker_name_from_drawing_ctx (draw_ctx, MarkerType::Start).is_null ()
+     || !get_marker_name_from_drawing_ctx (draw_ctx, MarkerType::Middle).is_null ()
+     || !get_marker_name_from_drawing_ctx (draw_ctx, MarkerType::End).is_null ())
 }
 
-fn render_marker_at_start_of_segment (segment: &Segment,
-                                      marker_name: *const libc::c_char,
-                                      orient: f64,
-                                      linewidth: f64,
-                                      ctx: *mut RsvgDrawingCtx) {
-    let xpos: f64;
-    let ypos: f64;
-
-    match *segment {
-        Segment::Degenerate { x, y } => {
-            xpos = x;
-            ypos = y;
-        },
-
-        Segment::LineOrCurve { x1, y1, .. } => {
-            xpos = x1;
-            ypos = y1;
-        }
-    }
-
-    unsafe { rsvg_marker_render (marker_name, xpos, ypos, orient, linewidth, ctx); }
-}
-
-fn render_marker_at_end_of_segment (segment: &Segment,
-                                    marker_name: *const libc::c_char,
-                                    orient: f64,
-                                    linewidth: f64,
-                                    ctx: *mut RsvgDrawingCtx) {
-    let xpos: f64;
-    let ypos: f64;
-
-    match *segment {
-        Segment::Degenerate { x, y } => {
-            xpos = x;
-            ypos = y;
-        },
-
-        Segment::LineOrCurve { x4, y4, .. } => {
-            xpos = x4;
-            ypos = y4;
-        }
-    }
-
-    unsafe { rsvg_marker_render (marker_name, xpos, ypos, orient, linewidth, ctx); }
-}
-
-#[no_mangle]
-pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
-                                   raw_builder: *mut RsvgPathBuilder) {
-    let linewidth: f64 = unsafe { rsvg_get_normalized_stroke_width (ctx) };
+pub fn render_markers_for_path_builder (builder:  &RsvgPathBuilder,
+                                        draw_ctx: *const RsvgDrawingCtx) {
+    let linewidth: f64 = unsafe { rsvg_get_normalized_stroke_width (draw_ctx) };
 
     if linewidth == 0.0 {
         return;
     }
 
-    let start_marker = unsafe { rsvg_get_start_marker (ctx) };
-    let middle_marker = unsafe { rsvg_get_middle_marker (ctx) };
-    let end_marker = unsafe { rsvg_get_end_marker (ctx) };
-
-    if start_marker.is_null () && middle_marker.is_null () && end_marker.is_null () {
+    if !drawing_ctx_has_markers (draw_ctx) {
         return;
     }
 
-    assert! (!raw_builder.is_null ());
-    let builder: &mut RsvgPathBuilder = unsafe { &mut (*raw_builder) };
+    emit_markers_for_path_builder (builder,
+                                   &mut |marker_type: MarkerType, x: f64, y: f64, computed_angle: f64| {
+                                       emit_marker_by_name (draw_ctx,
+                                                            get_marker_name_from_drawing_ctx (draw_ctx, marker_type),
+                                                            x,
+                                                            y,
+                                                            computed_angle,
+                                                            linewidth);
+                                   });
+}
+
+fn emit_markers_for_path_builder<E> (builder: &RsvgPathBuilder,
+                                     emit_fn: &mut E) where E: FnMut(MarkerType, f64, f64, f64) {
+    enum SubpathState {
+        NoSubpath,
+        InSubpath
+    };
 
     /* Convert the path to a list of segments and bare points */
     let segments = path_builder_to_segments (builder);
@@ -416,14 +676,18 @@ pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
                         /* Got a lone point after a subpath; render the subpath's end marker first */
 
                         let (_, incoming_vx, incoming_vy) = find_incoming_directionality_backwards (&segments, i - 1);
-                        render_marker_at_end_of_segment (&segments[i - 1], end_marker, angle_from_vector (incoming_vx, incoming_vy), linewidth, ctx);
+                        emit_marker (&segments[i - 1],
+                                     MarkerEndpoint::End,
+                                     MarkerType::End,
+                                     angle_from_vector (incoming_vx, incoming_vy),
+                                     emit_fn);
                     },
 
                     _ => { }
                 }
 
                 /* Render marker for the lone point; no directionality */
-                render_marker_at_start_of_segment (segment, middle_marker, 0.0, linewidth, ctx);
+                emit_marker (segment, MarkerEndpoint::Start, MarkerType::Middle, 0.0, emit_fn);
 
                 subpath_state = SubpathState::NoSubpath;
             },
@@ -434,7 +698,11 @@ pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
                 match subpath_state {
                     SubpathState::NoSubpath => {
                         let (_, outgoing_vx, outgoing_vy) = find_outgoing_directionality_forwards (&segments, i);
-                        render_marker_at_start_of_segment (segment, start_marker, angle_from_vector (outgoing_vx, outgoing_vy), linewidth, ctx);
+                        emit_marker (segment,
+                                     MarkerEndpoint::Start,
+                                     MarkerType::Start,
+                                     angle_from_vector (outgoing_vx, outgoing_vy),
+                                     emit_fn);
 
                         subpath_state = SubpathState::InSubpath;
                     },
@@ -454,7 +722,7 @@ pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
                         let angle: f64;
 
                         if has_incoming && has_outgoing {
-                            angle = (incoming + outgoing) / 2.0;
+                            angle = bisect_angles (incoming, outgoing);
                         } else if has_incoming {
                             angle = incoming;
                         } else if has_outgoing {
@@ -463,7 +731,7 @@ pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
                             angle = 0.0;
                         }
 
-                        render_marker_at_start_of_segment (segment, middle_marker, angle, linewidth, ctx);
+                        emit_marker (segment, MarkerEndpoint::Start, MarkerType::Middle, angle, emit_fn);
                     }
                 }
             }
@@ -478,7 +746,7 @@ pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
             Segment::LineOrCurve { .. } => {
                 let (_, incoming_vx, incoming_vy) = find_incoming_directionality_backwards (&segments, segments.len () - 1);
 
-                render_marker_at_end_of_segment (&segment, end_marker, angle_from_vector (incoming_vx, incoming_vy), linewidth, ctx);
+                emit_marker (&segment, MarkerEndpoint::End, MarkerType::End, angle_from_vector (incoming_vx, incoming_vy), emit_fn);
             },
 
             _ => { }
@@ -486,15 +754,124 @@ pub extern fn rsvg_render_markers (ctx: *mut RsvgDrawingCtx,
     }
 }
 
-
 /******************** Tests ********************/
 
 #[cfg(test)]
-mod tests {
+mod parser_tests {
+    use super::*;
+
+    #[test]
+    fn parsing_invalid_marker_units_yields_error () {
+        assert! (is_parse_error (&MarkerUnits::from_str ("").map_err (|e| AttributeError::from (e))));
+        assert! (is_parse_error (&MarkerUnits::from_str ("foo").map_err (|e| AttributeError::from (e))));
+    }
+
+    #[test]
+    fn parses_marker_units () {
+        assert_eq! (MarkerUnits::from_str ("userSpaceOnUse"), Ok (MarkerUnits::UserSpaceOnUse));
+        assert_eq! (MarkerUnits::from_str ("strokeWidth"),    Ok (MarkerUnits::StrokeWidth));
+    }
+
+    #[test]
+    fn parsing_invalid_marker_orient_yields_error () {
+        assert! (is_parse_error (&MarkerOrient::from_str ("").map_err (|e| AttributeError::from (e))));
+        assert! (is_parse_error (&MarkerOrient::from_str ("blah").map_err (|e| AttributeError::from (e))));
+        assert! (is_parse_error (&MarkerOrient::from_str ("45blah").map_err (|e| AttributeError::from (e))));
+    }
+
+    #[test]
+    fn parses_marker_orient () {
+        assert_eq! (MarkerOrient::from_str ("auto"), Ok (MarkerOrient::Auto));
+
+        assert_eq! (MarkerOrient::from_str ("0"), Ok (MarkerOrient::Degrees (0.0)));
+        assert_eq! (MarkerOrient::from_str ("180"), Ok (MarkerOrient::Degrees (180.0)));
+        assert_eq! (MarkerOrient::from_str ("180deg"), Ok (MarkerOrient::Degrees (180.0)));
+        assert_eq! (MarkerOrient::from_str ("-400grad"), Ok (MarkerOrient::Degrees (-360.0)));
+        assert_eq! (MarkerOrient::from_str ("1rad"), Ok (MarkerOrient::Degrees (180.0 / PI)));
+    }
+}
+
+#[cfg(test)]
+mod directionality_tests {
+    use std::f64::consts::*;
     use super::*;
-    use path_builder::*;
     extern crate cairo;
 
+    fn test_bisection_angle (expected: f64,
+                             incoming_vx: f64,
+                             incoming_vy: f64,
+                             outgoing_vx: f64,
+                             outgoing_vy: f64) {
+        let bisected = super::bisect_angles (super::angle_from_vector (incoming_vx, incoming_vy),
+                                             super::angle_from_vector (outgoing_vx, outgoing_vy));
+        assert! (double_equals (expected, bisected));
+    }
+
+    #[test]
+    fn bisection_angle_is_correct_from_incoming_counterclockwise_to_outgoing () {
+        // 1st quadrant
+        test_bisection_angle (FRAC_PI_4,
+                              1.0, 0.0,
+                              0.0, 1.0);
+
+        // 2nd quadrant
+        test_bisection_angle (FRAC_PI_2 + FRAC_PI_4,
+                              0.0, 1.0,
+                              -1.0, 0.0);
+
+        // 3rd quadrant
+        test_bisection_angle (PI + FRAC_PI_4,
+                              -1.0, 0.0,
+                              0.0, -1.0);
+
+        // 4th quadrant
+        test_bisection_angle (PI + FRAC_PI_2 + FRAC_PI_4,
+                              0.0, -1.0,
+                              1.0, 0.0);
+    }
+
+    #[test]
+    fn bisection_angle_is_correct_from_incoming_clockwise_to_outgoing () {
+        // 1st quadrant
+        test_bisection_angle (FRAC_PI_4,
+                              0.0, 1.0,
+                              1.0, 0.0);
+
+        // 2nd quadrant
+        test_bisection_angle (FRAC_PI_2 + FRAC_PI_4,
+                              -1.0, 0.0,
+                              0.0, 1.0);
+
+        // 3rd quadrant
+        test_bisection_angle (PI + FRAC_PI_4,
+                              0.0, -1.0,
+                              -1.0, 0.0);
+
+        // 4th quadrant
+        test_bisection_angle (PI + FRAC_PI_2 + FRAC_PI_4,
+                              1.0, 0.0,
+                              0.0, -1.0);
+    }
+
+    #[test]
+    fn bisection_angle_is_correct_for_more_than_quarter_turn_angle () {
+        test_bisection_angle (0.0,
+                              0.1, -1.0,
+                              0.1, 1.0);
+
+        test_bisection_angle (FRAC_PI_2,
+                              1.0, 0.1,
+                              -1.0, 0.1);
+
+        test_bisection_angle (PI,
+                              -0.1, 1.0,
+                              -0.1, -1.0);
+
+        test_bisection_angle (PI + FRAC_PI_2,
+                              -1.0, -0.1,
+                              1.0, -0.1);
+    }
+
     fn degenerate (x: f64, y: f64) -> Segment {
         super::make_degenerate (x, y)
     }
@@ -658,7 +1035,7 @@ mod tests {
 
     /* Sequence of moveto; should generate degenerate points.
      *
-     * This test is not enabled right now!  We create the 
+     * This test is not enabled right now!  We create the
      * path fixtures with Cairo, and Cairo compresses
      * sequences of moveto into a single one.  So, we can't
      * really test this, as we don't get the fixture we want.
@@ -719,17 +1096,17 @@ mod tests {
 
     #[test]
     fn curves_with_loops_and_coincident_ends_have_directionality () {
-        let (v1x, v1y, v2x, v2y) = 
+        let (v1x, v1y, v2x, v2y) =
             super::get_segment_directionalities (&curve (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 1.0, 2.0)).unwrap ();
         assert_eq! ((2.0, 2.0), (v1x, v1y));
         assert_eq! ((-4.0, -4.0), (v2x, v2y));
 
-        let (v1x, v1y, v2x, v2y) = 
+        let (v1x, v1y, v2x, v2y) =
             super::get_segment_directionalities (&curve (1.0, 2.0, 1.0, 2.0, 3.0, 4.0, 1.0, 2.0)).unwrap ();
         assert_eq! ((2.0, 2.0), (v1x, v1y));
         assert_eq! ((-2.0, -2.0), (v2x, v2y));
 
-        let (v1x, v1y, v2x, v2y) = 
+        let (v1x, v1y, v2x, v2y) =
             super::get_segment_directionalities (&curve (1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 1.0, 2.0)).unwrap ();
         assert_eq! ((2.0, 2.0), (v1x, v1y));
         assert_eq! ((-2.0, -2.0), (v2x, v2y));
@@ -766,3 +1143,55 @@ mod tests {
         assert_eq! ((40.0, 30.0), (v2x, v2y));
     }
 }
+
+#[cfg(test)]
+mod marker_tests {
+    use super::*;
+    extern crate cairo;
+
+    #[test]
+    fn emits_for_open_subpath () {
+        let mut builder = RsvgPathBuilder::new ();
+        builder.move_to (0.0, 0.0);
+        builder.line_to (1.0, 0.0);
+        builder.line_to (1.0, 1.0);
+        builder.line_to (0.0, 1.0);
+
+        let mut v = Vec::new ();
+
+        emit_markers_for_path_builder (&builder,
+                                       &mut |marker_type: MarkerType, x: f64, y: f64, computed_angle: f64| {
+                                           v.push ((marker_type, x, y, computed_angle));
+                                       });
+
+        assert_eq! (v, vec! [(MarkerType::Start,  0.0, 0.0, 0.0),
+                             (MarkerType::Middle, 1.0, 0.0, angle_from_vector (1.0, 1.0)),
+                             (MarkerType::Middle, 1.0, 1.0, angle_from_vector (-1.0, 1.0)),
+                             (MarkerType::End,    0.0, 1.0, angle_from_vector (-1.0, 0.0))]);
+    }
+
+    #[test]
+    #[ignore]
+    // https://bugzilla.gnome.org/show_bug.cgi?id=777854
+    fn emits_for_closed_subpath () {
+        let mut builder = RsvgPathBuilder::new ();
+        builder.move_to (0.0, 0.0);
+        builder.line_to (1.0, 0.0);
+        builder.line_to (1.0, 1.0);
+        builder.line_to (0.0, 1.0);
+        builder.close_path ();
+
+        let mut v = Vec::new ();
+
+        emit_markers_for_path_builder (&builder,
+                                       &mut |marker_type: MarkerType, x: f64, y: f64, computed_angle: f64| {
+                                           v.push ((marker_type, x, y, computed_angle));
+                                       });
+
+        assert_eq! (v, vec! [(MarkerType::Start,  0.0, 0.0, 0.0),
+                             (MarkerType::Middle, 1.0, 0.0, angle_from_vector (1.0, 1.0)),
+                             (MarkerType::Middle, 1.0, 1.0, angle_from_vector (-1.0, 1.0)),
+                             (MarkerType::Middle, 0.0, 1.0, angle_from_vector (-1.0, 0.0)),
+                             (MarkerType::End,    0.0, 0.0, angle_from_vector (1.0, -1.0))]);
+    }
+}
