@@ -6,16 +6,14 @@ use std::str::Chars;
 use std::iter::Enumerate;
 use path_builder::*;
 
-use self::glib::translate::*;
-
 extern crate cairo;
 
-struct PathParser<'external> {
-    chars_enumerator: Enumerate<Chars<'external>>,
+struct PathParser<'b> {
+    chars_enumerator: Enumerate<Chars<'b>>,
     lookahead: Option <char>, /* None if we are in EOF */
     current_pos: usize,
 
-    builder: &'external mut RsvgPathBuilder,
+    builder: &'b mut RsvgPathBuilder,
 
     error_message: &'static str,
     has_error: bool,
@@ -68,8 +66,8 @@ struct PathParser<'external> {
  *     M.1-2,3E2-4
  *     M 0.1 -2 300 -4
  */
-impl<'external> PathParser<'external> {
-    fn new (builder: &'external mut RsvgPathBuilder, path_str: &'external str) -> PathParser<'external> {
+impl<'b> PathParser<'b> {
+    fn new (builder: &'b mut RsvgPathBuilder, path_str: &'b str) -> PathParser<'b> {
         PathParser {
             chars_enumerator: path_str.chars ().enumerate (),
             lookahead: None,
@@ -371,7 +369,7 @@ impl<'external> PathParser<'external> {
         self.builder.curve_to (x2, y2, x3, y3, x4, y4);
     }
 
-    fn emit_arc (&mut self, rx: f64, ry: f64, x_axis_rotation: f64, large_arc: bool, sweep: bool, x: f64, y: f64) {
+    fn emit_arc (&mut self, rx: f64, ry: f64, x_axis_rotation: f64, large_arc: LargeArc, sweep: Sweep, x: f64, y: f64) {
         let (start_x, start_y) = (self.current_x, self.current_y);
 
         self.set_current_point (x, y);
@@ -907,16 +905,23 @@ impl<'external> PathParser<'external> {
                         if let Some (large_arc_flag) = self.flag () {
                             assert! (self.optional_comma_whitespace ());
 
+                            let large_arc = LargeArc (large_arc_flag);
+
                             if let Some (sweep_flag) = self.flag () {
                                 assert! (self.optional_comma_whitespace ());
 
+                                let sweep = match sweep_flag {
+                                    false => Sweep::Negative,
+                                    true => Sweep::Positive
+                                };
+
                                 if let Some ((mut x, mut y)) = self.coordinate_pair () {
                                     if !absolute {
                                         x += self.current_x;
                                         y += self.current_y;
                                     }
 
-                                    self.emit_arc (rx, ry, x_axis_rotation, large_arc_flag, sweep_flag, x, y);
+                                    self.emit_arc (rx, ry, x_axis_rotation, large_arc, sweep, x, y);
 
                                     self.whitespace ();
 
@@ -999,91 +1004,12 @@ pub fn parse_path_into_builder (path_str: &str, builder: &mut RsvgPathBuilder) -
     }
 }
 
-#[no_mangle]
-pub extern fn rsvg_path_parser_from_str_into_builder (path_str: *const libc::c_char) -> *mut RsvgPathBuilder {
-    let mut builder = RsvgPathBuilder::new ();
-    let my_path_str = unsafe { &String::from_glib_none (path_str) };
-
-    {
-        let mut parser = PathParser::new (&mut builder, my_path_str);
-        parser.parse ();
-        /* FIXME: we aren't passing errors back to the caller. */
-    }
-
-    let boxed_builder = Box::new (builder);
-
-    Box::into_raw (boxed_builder)
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
-    use path_builder::*;
 
     extern crate cairo;
 
-    fn path_segment_vectors_are_equal (a: &Vec<cairo::PathSegment>,
-                                       b: &Vec<cairo::PathSegment>) -> bool {
-
-        if a.len() != b.len () {
-            return false;
-        }
-
-        if a.len () == 0 && b.len () == 0 {
-            return true;
-        }
-
-        let mut iter = a.iter().zip (b);
-
-        loop {
-            if let Some ((seg1, seg2)) = iter.next () {
-                match *seg1 {
-                    cairo::PathSegment::MoveTo ((x, y)) => {
-                        if let cairo::PathSegment::MoveTo ((ox, oy)) = *seg2 {
-                            if (x, y) != (ox, oy) {
-                                return false;
-                            }
-                        } else {
-                            return false;
-                        }
-                    },
-
-                    cairo::PathSegment::LineTo ((x, y)) => {
-                        if let cairo::PathSegment::LineTo ((ox, oy)) = *seg2 {
-                            if (x, y) != (ox, oy) {
-                                return false;
-                            }
-                        } else {
-                            return false;
-                        }
-                    },
-
-                    cairo::PathSegment::CurveTo ((x2, y2), (x3, y3), (x4, y4)) => {
-                        if let cairo::PathSegment::CurveTo ((ox2, oy2), (ox3, oy3), (ox4, oy4)) = *seg2 {
-                            if (ox2, oy2, ox3, oy3, ox4, oy4) != (x2, y2, x3, y3, x4, y4) {
-                                return false;
-                            }
-                        } else {
-                            return false;
-                        }
-                    },
-
-                    cairo::PathSegment::ClosePath => {
-                        if let cairo::PathSegment::ClosePath = *seg2 {
-                            /* okay */
-                        } else {
-                            return false;
-                        }
-                    }
-                }
-            } else {
-                break;
-            }
-        }
-
-        true
-    }
-
     fn print_error (error: &ParseError, path_str: &str) {
         let prefix = "Error in \"";
 
@@ -1111,33 +1037,33 @@ mod tests {
     }
 
     fn test_parser (path_str: &str,
-                    expected_segments: &Vec<cairo::PathSegment>) {
+                    expected_commands: &[PathCommand]) {
         let builder = parse_path (path_str);
-        let segments = builder.get_path_segments ();
+        let commands = builder.get_path_commands ();
 
-        assert! (path_segment_vectors_are_equal (expected_segments, segments));
+        assert_eq! (expected_commands, commands);
     }
 
-    fn moveto (x: f64, y: f64) -> cairo::PathSegment {
-        cairo::PathSegment::MoveTo ((x, y))
+    fn moveto (x: f64, y: f64) -> PathCommand {
+        PathCommand::MoveTo (x, y)
     }
 
-    fn lineto (x: f64, y: f64) -> cairo::PathSegment {
-        cairo::PathSegment::LineTo ((x, y))
+    fn lineto (x: f64, y: f64) -> PathCommand {
+        PathCommand::LineTo (x, y)
     }
 
-    fn curveto (x2: f64, y2: f64, x3: f64, y3: f64, x4: f64, y4: f64) -> cairo::PathSegment {
-        cairo::PathSegment::CurveTo ((x2, y2), (x3, y3), (x4, y4))
+    fn curveto (x2: f64, y2: f64, x3: f64, y3: f64, x4: f64, y4: f64) -> PathCommand {
+        PathCommand::CurveTo ((x2, y2), (x3, y3), (x4, y4))
     }
 
-    fn closepath () -> cairo::PathSegment {
-        cairo::PathSegment::ClosePath
+    fn closepath () -> PathCommand {
+        PathCommand::ClosePath
     }
 
     #[test]
     fn handles_empty_data () {
         test_parser ("",
-                     &Vec::<cairo::PathSegment>::new ());
+                     &Vec::<PathCommand>::new ());
     }
 
     #[test]
